import pytest

from blends.models import SyntaxGraph
from blends.stack.attributes import (
    STACK_GRAPH_IS_DEFINITION,
    STACK_GRAPH_IS_REFERENCE,
    STACK_GRAPH_KIND,
    STACK_GRAPH_SYMBOL,
)
from blends.stack.node_kinds import StackGraphNodeKind
from blends.stack.partial_path import (
    PartialPathLimits,
    compute_minimal_partial_paths_in_file,
)
from blends.stack.view import StackGraphView


def _view_for_graph(graph: SyntaxGraph) -> StackGraphView:
    return StackGraphView.from_syntax_graph(graph, path="test.py")


@pytest.mark.blends_test_group("stack_unittesting")
def test_minimal_paths_seed_count_includes_root_and_jump() -> None:
    graph = SyntaxGraph()
    graph.add_node("1")
    graph.update_node_attribute("1", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("1", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("1", STACK_GRAPH_IS_REFERENCE, value=True)

    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("2", STACK_GRAPH_IS_DEFINITION, value=True)

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.JUMP_TO.value)

    view = _view_for_graph(graph)
    result = compute_minimal_partial_paths_in_file(view, "test.py", limits=PartialPathLimits())
    expected_seed_count = 4

    assert result.stats.seed_count == expected_seed_count


@pytest.mark.blends_test_group("stack_unittesting")
def test_minimal_paths_accept_reference_to_definition() -> None:
    graph = SyntaxGraph()
    graph.add_node("1")
    graph.update_node_attribute("1", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("1", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("1", STACK_GRAPH_IS_REFERENCE, value=True)

    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("2", STACK_GRAPH_IS_DEFINITION, value=True)

    graph.add_edge("1", "2", precedence=0)

    view = _view_for_graph(graph)
    result = compute_minimal_partial_paths_in_file(view, "test.py", limits=PartialPathLimits())

    push_index = view.nid_to_index["1"]
    pop_index = view.nid_to_index["2"]
    accepted_pairs = {(path.start_node_index, path.end_node_index) for path in result.paths}

    assert (push_index, pop_index) in accepted_pairs


@pytest.mark.blends_test_group("stack_unittesting")
def test_minimal_paths_accept_jump_boundary() -> None:
    graph = SyntaxGraph()
    graph.add_node("1")
    graph.update_node_attribute("1", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("1", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("1", STACK_GRAPH_IS_REFERENCE, value=True)

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.JUMP_TO.value)

    graph.add_edge("1", "3", precedence=0)

    view = _view_for_graph(graph)
    result = compute_minimal_partial_paths_in_file(view, "test.py", limits=PartialPathLimits())

    jump_index = view.nid_to_index["3"]
    assert any(path.end_node_index == jump_index for path in result.paths)


@pytest.mark.blends_test_group("stack_unittesting")
def test_minimal_paths_forward_expands_past_non_endpoint() -> None:
    graph = SyntaxGraph()
    graph.add_node("1")
    graph.update_node_attribute("1", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("1", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("1", STACK_GRAPH_IS_REFERENCE, value=True)

    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("3", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("3", STACK_GRAPH_IS_DEFINITION, value=True)

    graph.add_edge("1", "2", precedence=0)
    graph.add_edge("2", "3", precedence=0)

    view = _view_for_graph(graph)
    result = compute_minimal_partial_paths_in_file(view, "test.py", limits=PartialPathLimits())

    push_index = view.nid_to_index["1"]
    pop_index = view.nid_to_index["3"]
    accepted_pairs = {(path.start_node_index, path.end_node_index) for path in result.paths}

    assert (push_index, pop_index) in accepted_pairs
    assert result.stats.length_histogram.get(2) == 1



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.do_statement import (
    build_do_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    body_id = graph.nodes[args.n_id]["label_field_body"]
    if graph.nodes[body_id]["label_type"] == "expression_statement":
        body_id = match_ast(graph, body_id)["__0__"]

    condition_node = graph.nodes[args.n_id]["label_field_condition"]

    return build_do_statement_node(args, body_id, condition_node)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
    match_ast_d,
    pred_ast,
)
from blends.syntax.builders.export_statement import (
    build_export_statement_node,
)
from blends.syntax.builders.import_global import (
    build_import_global_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    nodes = graph.nodes
    alias: str | None = None

    target_n_ids = match_ast(graph, args.n_id, "string_literal", "as", "library_export", depth=-1)

    exp_n_id = target_n_ids["string_literal"]
    alias_n_id = target_n_ids["as"]
    expression = ""
    if exp_n_id is not None:
        expression = nodes[exp_n_id].get("label_text", "")[1:-1]

    if alias_n_id:
        p_id = pred_ast(graph, alias_n_id)[0]
        identifier_n_id = match_ast_d(graph, p_id, "identifier")
        alias = ""
        if identifier_n_id is not None:
            alias = nodes[identifier_n_id].get("label_text")

    if target_n_ids["library_export"]:
        return build_export_statement_node(args, expression, None)

    return build_import_global_node(args, expression, set(), alias)



from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING

from blends.stack.attributes import (
    STACK_GRAPH_IMPORT_MODULE_STEM,
    STACK_GRAPH_IMPORT_SOURCE_SYMBOL,
    STACK_GRAPH_IS_DEFINITION,
    STACK_GRAPH_IS_EXPORTED,
    STACK_GRAPH_IS_IMPORT_BINDING,
    STACK_GRAPH_IS_REFERENCE,
    STACK_GRAPH_KIND,
    STACK_GRAPH_PRECEDENCE,
    STACK_GRAPH_SCOPE,
    STACK_GRAPH_SCOPE_TYPE,
    STACK_GRAPH_SYMBOL,
)
from blends.stack.node_kinds import (
    StackGraphNodeKind,
)

if TYPE_CHECKING:
    from blends.models import (
        NId,
        SyntaxGraph,
    )

ROOT_INDEX: int = 0
JUMP_TO_INDEX: int = 1


class Degree(IntEnum):
    ZERO = 0
    ONE = 1
    MULTIPLE = 2


def _nid_sort_key(n_id: NId) -> tuple[int, int | str]:
    if n_id.isdecimal():
        return (0, int(n_id))
    return (1, n_id)


def _resolve_file_path(graph: SyntaxGraph, path: str | None) -> str:
    if path is not None:
        return path
    metadata = graph.nodes.get("0", {})
    node_path = metadata.get("path")
    return node_path if isinstance(node_path, str) else ""


def _collect_stack_graph_node_ids(graph: SyntaxGraph) -> list[NId]:
    return [
        node_id for node_id in graph.nodes if graph.nodes[node_id].get(STACK_GRAPH_KIND) is not None
    ]


@dataclass(slots=True)
class NodeSemantics:
    node_kind: list[str | None]
    node_symbol_id: list[int | None]
    node_scope_index: list[int | None]
    node_precedence: list[int]
    exported_scopes: set[int]
    node_is_reference: list[bool]
    node_is_definition: list[bool]
    node_scope_type: list[str | None]
    node_is_import_binding: list[bool]
    node_import_module_stem: list[str | None]
    node_import_source_symbol: list[str | None]


def _find_root_nid(graph: SyntaxGraph, stack_node_ids: list[NId]) -> NId | None:
    for nid in stack_node_ids:
        if graph.nodes[nid].get(STACK_GRAPH_KIND) == StackGraphNodeKind.ROOT.value:
            return nid
    return None


def _build_node_index_maps(
    stack_node_ids: list[NId], *, root_nid: NId | None
) -> tuple[dict[NId, int], list[NId]]:
    nid_to_index: dict[NId, int] = {}
    index_to_nid: list[NId] = []

    if root_nid is not None:
        nid_to_index[root_nid] = ROOT_INDEX
    index_to_nid.append(root_nid or "")

    index_to_nid.append("")

    for node_id in stack_node_ids:
        if node_id in nid_to_index:
            continue
        nid_to_index[node_id] = len(index_to_nid)
        index_to_nid.append(node_id)
    return nid_to_index, index_to_nid


def _fill_structural_attrs(
    attrs: dict[str, object],
    idx: int,
    nid_to_index: dict[NId, int],
    symbols: SymbolInterner,
    sem: NodeSemantics,
) -> None:
    kind = attrs.get(STACK_GRAPH_KIND)
    if isinstance(kind, str):
        sem.node_kind[idx] = kind

    symbol = attrs.get(STACK_GRAPH_SYMBOL)
    if isinstance(symbol, str) and symbol:
        sem.node_symbol_id[idx] = symbols.intern(symbol)

    scope = attrs.get(STACK_GRAPH_SCOPE)
    if isinstance(scope, str):
        sem.node_scope_index[idx] = nid_to_index.get(scope)

    precedence = attrs.get(STACK_GRAPH_PRECEDENCE, 0)
    sem.node_precedence[idx] = precedence if isinstance(precedence, int) else 0


def _fill_flag_attrs(
    attrs: dict[str, object],
    idx: int,
    sem: NodeSemantics,
) -> None:
    if attrs.get(STACK_GRAPH_IS_EXPORTED) is True:
        sem.exported_scopes.add(idx)

    is_reference = attrs.get(STACK_GRAPH_IS_REFERENCE)
    if isinstance(is_reference, bool):
        sem.node_is_reference[idx] = is_reference

    is_definition = attrs.get(STACK_GRAPH_IS_DEFINITION)
    if isinstance(is_definition, bool):
        sem.node_is_definition[idx] = is_definition

    scope_type = attrs.get(STACK_GRAPH_SCOPE_TYPE)
    if isinstance(scope_type, str):
        sem.node_scope_type[idx] = scope_type

    is_import_binding = attrs.get(STACK_GRAPH_IS_IMPORT_BINDING)
    if isinstance(is_import_binding, bool):
        sem.node_is_import_binding[idx] = is_import_binding

    import_module_stem = attrs.get(STACK_GRAPH_IMPORT_MODULE_STEM)
    if isinstance(import_module_stem, str):
        sem.node_import_module_stem[idx] = import_module_stem

    import_source_symbol = attrs.get(STACK_GRAPH_IMPORT_SOURCE_SYMBOL)
    if isinstance(import_source_symbol, str):
        sem.node_import_source_symbol[idx] = import_source_symbol


def _extract_node_semantics(
    graph: SyntaxGraph,
    *,
    stack_node_ids: list[NId],
    nid_to_index: dict[NId, int],
    symbols: SymbolInterner,
    node_count: int,
) -> NodeSemantics:
    sem = NodeSemantics(
        node_kind=[None] * node_count,
        node_symbol_id=[None] * node_count,
        node_scope_index=[None] * node_count,
        node_precedence=[0] * node_count,
        exported_scopes=set(),
        node_is_reference=[False] * node_count,
        node_is_definition=[False] * node_count,
        node_scope_type=[None] * node_count,
        node_is_import_binding=[False] * node_count,
        node_import_module_stem=[None] * node_count,
        node_import_source_symbol=[None] * node_count,
    )
    sem.node_kind[ROOT_INDEX] = StackGraphNodeKind.ROOT.value
    sem.node_kind[JUMP_TO_INDEX] = StackGraphNodeKind.JUMP_TO.value

    for node_id in stack_node_ids:
        attrs = graph.nodes[node_id]
        idx = nid_to_index[node_id]
        _fill_structural_attrs(attrs, idx, nid_to_index, symbols, sem)
        _fill_flag_attrs(attrs, idx, sem)

    return sem


def _extract_and_normalize_outgoing(
    graph: SyntaxGraph,
    *,
    stack_node_ids: list[NId],
    nid_to_index: dict[NId, int],
    node_count: int,
) -> tuple[list[list[tuple[int, int]]], list[Degree]]:
    outgoing: list[list[tuple[int, int]]] = [[] for _ in range(node_count)]
    incoming_counts: list[int] = [0] * node_count
    stack_node_id_set = set(stack_node_ids)
    for source_id, sink_id, edge_attrs in graph.edges(data=True):
        if source_id not in stack_node_id_set or sink_id not in stack_node_id_set:
            continue
        if "precedence" not in edge_attrs:
            continue
        source_index = nid_to_index[source_id]
        sink_index = nid_to_index[sink_id]
        precedence = edge_attrs.get("precedence", 0)
        outgoing[source_index].append(
            (
                sink_index,
                precedence if isinstance(precedence, int) else 0,
            )
        )
        incoming_counts[sink_index] += 1

    normalized_outgoing: list[list[tuple[int, int]]] = [[] for _ in range(node_count)]
    for source_index, edges in enumerate(outgoing):
        if not edges:
            continue
        edges.sort(key=lambda item: (item[0], item[1]))
        normalized_outgoing[source_index] = edges

    incoming_degree = [
        Degree.ZERO if count == 0 else Degree.ONE if count == 1 else Degree.MULTIPLE
        for count in incoming_counts
    ]
    return normalized_outgoing, incoming_degree


def _build_file_maps(
    file_path: str, *, stack_node_ids: list[NId], nid_to_index: dict[NId, int], node_count: int
) -> tuple[dict[str, frozenset[int]], list[str]]:
    real_node_indices = frozenset(nid_to_index[node_id] for node_id in stack_node_ids)
    file_to_nodes = {file_path: real_node_indices}
    node_to_file = [file_path if index in real_node_indices else "" for index in range(node_count)]
    return file_to_nodes, node_to_file


@dataclass(slots=True)
class SymbolInterner:
    symbol_to_id: dict[str, int]
    id_to_symbol: list[str]

    def intern(self, symbol: str) -> int:
        symbol_id = self.symbol_to_id.get(symbol)
        if symbol_id is not None:
            return symbol_id
        symbol_id = len(self.id_to_symbol)
        self.symbol_to_id[symbol] = symbol_id
        self.id_to_symbol.append(symbol)
        return symbol_id

    def lookup(self, symbol_id: int) -> str:
        return self.id_to_symbol[symbol_id]


@dataclass(frozen=True, slots=True)
class StackGraphView:
    file_path: str
    nid_to_index: dict[NId, int]
    index_to_nid: list[NId]
    symbols: SymbolInterner
    node_kind: list[str | None]
    node_symbol_id: list[int | None]
    node_scope_index: list[int | None]
    node_precedence: list[int]
    outgoing: list[list[tuple[int, int]]]
    incoming_degree: list[Degree]
    exported_scopes: set[int]
    file_to_nodes: dict[str, frozenset[int]]
    node_to_file: list[str]
    node_is_reference: list[bool]
    node_is_definition: list[bool]
    node_scope_type: list[str | None]
    node_is_import_binding: list[bool]
    node_import_module_stem: list[str | None]
    node_import_source_symbol: list[str | None]

    @classmethod
    def from_syntax_graph(cls, graph: SyntaxGraph, *, path: str | None = None) -> StackGraphView:
        file_path = _resolve_file_path(graph, path)

        stack_node_ids = sorted(_collect_stack_graph_node_ids(graph), key=_nid_sort_key)
        root_nid = _find_root_nid(graph, stack_node_ids)
        nid_to_index, index_to_nid = _build_node_index_maps(stack_node_ids, root_nid=root_nid)

        symbols = SymbolInterner(symbol_to_id={}, id_to_symbol=[])

        node_count = len(index_to_nid)
        semantics = _extract_node_semantics(
            graph,
            stack_node_ids=stack_node_ids,
            nid_to_index=nid_to_index,
            symbols=symbols,
            node_count=node_count,
        )
        outgoing, incoming_degree = _extract_and_normalize_outgoing(
            graph,
            stack_node_ids=stack_node_ids,
            nid_to_index=nid_to_index,
            node_count=node_count,
        )
        file_to_nodes, node_to_file = _build_file_maps(
            file_path,
            stack_node_ids=stack_node_ids,
            nid_to_index=nid_to_index,
            node_count=node_count,
        )

        return cls(
            file_path=file_path,
            nid_to_index=nid_to_index,
            index_to_nid=index_to_nid,
            symbols=symbols,
            node_kind=semantics.node_kind,
            node_symbol_id=semantics.node_symbol_id,
            node_scope_index=semantics.node_scope_index,
            node_precedence=semantics.node_precedence,
            outgoing=outgoing,
            incoming_degree=incoming_degree,
            exported_scopes=semantics.exported_scopes,
            file_to_nodes=file_to_nodes,
            node_to_file=node_to_file,
            node_is_reference=semantics.node_is_reference,
            node_is_definition=semantics.node_is_definition,
            node_scope_type=semantics.node_scope_type,
            node_is_import_binding=semantics.node_is_import_binding,
            node_import_module_stem=semantics.node_import_module_stem,
            node_import_source_symbol=semantics.node_import_source_symbol,
        )

    def kind_at(self, node_index: int) -> str:
        if node_index == 0:
            return StackGraphNodeKind.ROOT.value
        if node_index == 1:
            return StackGraphNodeKind.JUMP_TO.value
        kind = self.node_kind[node_index]
        return kind if isinstance(kind, str) else ""

    def symbol_id_at(self, node_index: int) -> int | None:
        if node_index in {0, 1}:
            return None
        return self.node_symbol_id[node_index]

    def scope_index_at(self, node_index: int) -> int | None:
        if node_index in {0, 1}:
            return None
        return self.node_scope_index[node_index]



from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_comment_node(args: SyntaxGraphArgs, comment: str) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        comment=comment,
        label_type="Comment",
    )
    return args.n_id



from blends.symbolic_evaluation.dispatcher import (
    generic,
)
from blends.symbolic_evaluation.models import (
    SymbolicEvalArgs,
    SymbolicEvaluation,
)


def evaluate(args: SymbolicEvalArgs) -> SymbolicEvaluation:
    args.evaluation[args.n_id] = False
    if val_id := args.graph.nodes[args.n_id].get("value_id"):
        args.evaluation[args.n_id] = generic(args.fork_n_id(val_id)).danger

    if args.method_evaluators and (method_evaluator := args.method_evaluators.get("parameter")):
        args.evaluation[args.n_id] = method_evaluator(args).danger

    return SymbolicEvaluation(args.evaluation[args.n_id], args.triggers)



from blends.symbolic_evaluation.dispatcher import (
    generic,
)
from blends.symbolic_evaluation.models import (
    SymbolicEvalArgs,
    SymbolicEvaluation,
)


def evaluate(args: SymbolicEvalArgs) -> SymbolicEvaluation:
    args.evaluation[args.n_id] = False
    block_danger = False
    if args.n_id not in args.path:
        block_id = args.graph.nodes[args.n_id]["block_id"]
        block_danger = generic(args.fork_n_id(block_id)).danger

    declaration_danger = False
    if decl_id := args.graph.nodes[args.n_id].get("declaration_id"):
        declaration_danger = generic(args.fork_n_id(decl_id)).danger

    args.evaluation[args.n_id] = block_danger or declaration_danger

    return SymbolicEvaluation(args.evaluation[args.n_id], args.triggers)



from collections.abc import (
    Iterator,
)

from blends.symbolic_evaluation.context.search.model import (
    SearchArgs,
    SearchResult,
)


def search(args: SearchArgs) -> Iterator[SearchResult]:
    if args.symbol in args.graph.nodes[args.n_id].get("variable", "").split(","):
        yield True, args.n_id



from blends.models import (
    NId,
)
from blends.syntax.builders.while_statement import (
    build_while_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    condition_id = n_attrs["label_field_condition"]
    block = n_attrs["label_field_body"]

    return build_while_statement_node(args, block, condition_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.assignment import (
    build_assignment_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    alias_id = graph.nodes[args.n_id].get("label_field_alias")
    children = adj_ast(graph, args.n_id)
    var_id = adj_ast(graph, alias_id)[0] if alias_id else children[-1]
    return build_assignment_node(args, var_id, children[0], None)



from blends.stack.partial_path.bindings import (
    PartialScopeStackBindings,
    PartialSymbolStackBindings,
)
from blends.stack.partial_path.errors import (
    PartialPathResolutionError,
    PartialPathResolutionErrorCode,
)
from blends.stack.partial_path.minimal_paths import (
    PartialPathLimits,
    PerFilePartialPaths,
    PerFilePartialPathStats,
    compute_minimal_partial_paths_in_file,
)
from blends.stack.partial_path.partial_path import PartialPath
from blends.stack.partial_path.partial_stacks import (
    PartialScopedSymbol,
    PartialScopeStack,
    PartialSymbolStack,
)
from blends.stack.partial_path.variables import (
    ScopeStackVariable,
    SymbolStackVariable,
)
from blends.stack.partial_path_database import (
    FileHandle,
    PartialPathDatabase,
    PartialPathFileRecord,
    StackGraphViewMetadata,
    remap_symbol_id,
)
from blends.stack.partial_path_db import (
    PartialPathDb,
    PartialPathId,
)
from blends.stack.partial_path_indexer import (
    PartialPathIndexRequest,
    index_partial_paths_for_file,
)

__all__ = [
    "FileHandle",
    "PartialPath",
    "PartialPathDatabase",
    "PartialPathDb",
    "PartialPathFileRecord",
    "PartialPathId",
    "PartialPathIndexRequest",
    "PartialPathLimits",
    "PartialPathResolutionError",
    "PartialPathResolutionErrorCode",
    "PartialScopeStack",
    "PartialScopeStackBindings",
    "PartialScopedSymbol",
    "PartialSymbolStack",
    "PartialSymbolStackBindings",
    "PerFilePartialPathStats",
    "PerFilePartialPaths",
    "ScopeStackVariable",
    "StackGraphViewMetadata",
    "SymbolStackVariable",
    "compute_minimal_partial_paths_in_file",
    "index_partial_paths_for_file",
    "remap_symbol_id",
]



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.string_literal import (
    build_string_literal_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    value = node_to_str(graph, args.n_id)[1:-1]

    c_ids = (
        _id
        for _id in adj_ast(graph, args.n_id)
        if graph.nodes[_id].get("label_type") in {"variable_name", "subscript_expression"}
    )

    return build_string_literal_node(args, value, c_ids)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.parameter_list import (
    build_parameter_list_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    c_ids = (
        _id
        for _id in adj_ast(graph, args.n_id)
        if graph.nodes[_id]["label_type"] not in {"(", ")", ",", "[", "]"}
    )
    return build_parameter_list_node(args, c_ids=c_ids)



from blends.models import (
    NId,
)
from blends.syntax.builders.assignment import (
    build_assignment_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    n_attrs = args.ast_graph.nodes[args.n_id]
    var_id = n_attrs["label_field_name"]
    val_id = n_attrs["label_field_value"]
    return build_assignment_node(args, var_id, val_id, None)



import pytest

from blends.stack import (
    MultiFileResolutionRequest,
    resolve_definitions_multi_file,
)
from blends.stack.forward_stitcher import ForwardStitcherConfig
from test.stack.test_helpers import (
    QUERY_MAX_PHASES,
    QUERY_MAX_WORK,
    find_symbol_nodes,
    setup_multi_file_database,
)


@pytest.mark.blends_test_group("stack_unittesting")
def test_empty_result_for_non_reference_node() -> None:
    views, database = setup_multi_file_database({"cross_file_a.py": 1})
    view = views[1]

    result = resolve_definitions_multi_file(
        view,
        database,
        request=MultiFileResolutionRequest(file_handle=1, ref_node_index=0),
    )

    assert result == []


@pytest.mark.blends_test_group("stack_unittesting")
def test_empty_result_when_file_handle_absent() -> None:
    views, database = setup_multi_file_database({"case_closure.py": 1})
    view = views[1]
    symbol_id = view.symbols.symbol_to_id.get("captured")
    assert symbol_id is not None
    ref_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=False)
    assert ref_nodes

    result = resolve_definitions_multi_file(
        view,
        database,
        request=MultiFileResolutionRequest(file_handle=99, ref_node_index=ref_nodes[0]),
    )

    assert result == []


@pytest.mark.blends_test_group("stack_unittesting")
def test_stats_files_considered_counts_loaded_files() -> None:
    views, database = setup_multi_file_database({"case_closure.py": 1, "cross_file_a.py": 2})
    view = views[1]
    symbol_id = view.symbols.symbol_to_id.get("captured")
    assert symbol_id is not None
    ref_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=False)
    assert ref_nodes

    output = resolve_definitions_multi_file(
        view,
        database,
        request=MultiFileResolutionRequest(
            file_handle=1,
            ref_node_index=ref_nodes[0],
            include_stats=True,
            config=ForwardStitcherConfig(
                max_phases=QUERY_MAX_PHASES,
                max_total_work=QUERY_MAX_WORK,
            ),
        ),
    )
    assert isinstance(output, tuple)
    _result, stats = output

    assert stats.files_considered == 1


@pytest.mark.blends_test_group("stack_unittesting")
def test_cancelled_resolution_returns_cancelled_flag() -> None:
    views, database = setup_multi_file_database({"case_closure.py": 1})
    view = views[1]
    symbol_id = view.symbols.symbol_to_id.get("captured")
    assert symbol_id is not None
    ref_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=False)
    assert ref_nodes

    output = resolve_definitions_multi_file(
        view,
        database,
        request=MultiFileResolutionRequest(
            file_handle=1,
            ref_node_index=ref_nodes[0],
            include_stats=True,
            config=ForwardStitcherConfig(deadline=0.0),
        ),
    )
    assert isinstance(output, tuple)
    result, stats = output

    assert stats.cancelled is True
    assert result == []


@pytest.mark.blends_test_group("stack_unittesting")
def test_reference_resolves_to_source_definition() -> None:
    file_map = {"cross_file_c.py": 1, "cross_file_d.py": 2}
    views, database = setup_multi_file_database(file_map)
    fh_c = file_map["cross_file_c.py"]
    fh_d = file_map["cross_file_d.py"]
    view_d = views[fh_d]
    symbol_id = view_d.symbols.symbol_to_id.get("exported_value")
    assert symbol_id is not None
    ref_nodes = find_symbol_nodes(view_d, symbol_id=symbol_id, is_definition=False)
    assert ref_nodes, "cross_file_d must contain a reference to exported_value"

    result = resolve_definitions_multi_file(
        view_d,
        database,
        request=MultiFileResolutionRequest(
            file_handle=fh_d,
            ref_node_index=ref_nodes[0],
            config=ForwardStitcherConfig(
                max_phases=QUERY_MAX_PHASES,
                max_total_work=QUERY_MAX_WORK,
            ),
        ),
        views=views,
    )

    assert isinstance(result, list)
    assert len(result) >= 1, "exported_value must resolve to at least one definition"
    file_handles = [fh for fh, _ in result]
    assert fh_c in file_handles, (
        f"exported_value source is in cross_file_c (handle={fh_c}), got {file_handles}"
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_result_tuples_attribute_correct_file_handles() -> None:
    views, database = setup_multi_file_database({"case_closure.py": 1, "cross_file_a.py": 2})
    view = views[1]
    symbol_id = view.symbols.symbol_to_id.get("captured")
    assert symbol_id is not None
    ref_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=False)
    assert ref_nodes

    result = resolve_definitions_multi_file(
        view,
        database,
        request=MultiFileResolutionRequest(
            file_handle=1,
            ref_node_index=ref_nodes[0],
            config=ForwardStitcherConfig(
                max_phases=QUERY_MAX_PHASES,
                max_total_work=QUERY_MAX_WORK,
            ),
        ),
    )
    assert isinstance(result, list)
    assert len(result) >= 1, "Expected at least one definition of 'captured'"
    valid_handles = set(database.list_files())
    for fh, _idx in result:
        assert fh in valid_handles, f"File handle {fh} not in database {valid_handles}"


@pytest.mark.blends_test_group("stack_unittesting")
def test_javascript_reference_resolves_to_source_definition() -> None:
    file_map = {"cross_file_c.js": 1, "cross_file_d.js": 2}
    views, database = setup_multi_file_database(file_map)
    fh_c = file_map["cross_file_c.js"]
    fh_d = file_map["cross_file_d.js"]
    view_d = views[fh_d]
    symbol_id = view_d.symbols.symbol_to_id.get("exportedValue")
    assert symbol_id is not None
    ref_nodes = find_symbol_nodes(view_d, symbol_id=symbol_id, is_definition=False)
    assert ref_nodes, "cross_file_d.js must contain a reference to exportedValue"

    result = resolve_definitions_multi_file(
        view_d,
        database,
        request=MultiFileResolutionRequest(
            file_handle=fh_d,
            ref_node_index=ref_nodes[0],
            config=ForwardStitcherConfig(
                max_phases=QUERY_MAX_PHASES,
                max_total_work=QUERY_MAX_WORK,
            ),
        ),
        views=views,
    )

    assert isinstance(result, list)
    assert len(result) >= 1, "exportedValue must resolve to at least one definition"
    file_handles = [fh for fh, _ in result]
    assert fh_c in file_handles, (
        f"exportedValue source is in cross_file_c.js (handle={fh_c}), got {file_handles}"
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_sentinel_lazy_loads_only_direct_dependency() -> None:
    file_map = {"mf_core.py": 1, "mf_transport.py": 2}
    transport_fh = file_map["mf_transport.py"]
    views, database = setup_multi_file_database(file_map)
    view = views[transport_fh]
    symbol_id = view.symbols.symbol_to_id.get("SENTINEL")
    assert symbol_id is not None
    ref_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=False)
    assert ref_nodes

    output = resolve_definitions_multi_file(
        view,
        database,
        request=MultiFileResolutionRequest(
            file_handle=transport_fh,
            ref_node_index=ref_nodes[0],
            include_stats=True,
            config=ForwardStitcherConfig(
                max_phases=QUERY_MAX_PHASES,
                max_total_work=QUERY_MAX_WORK,
            ),
        ),
        views=views,
    )
    assert isinstance(output, tuple)
    _, stats = output
    assert stats.files_considered == len(file_map)


@pytest.mark.blends_test_group("stack_unittesting")
def test_unrelated_files_in_db_do_not_increase_files_loaded() -> None:
    all_files = {
        "mf_core.py": 1,
        "mf_transport.py": 2,
        "mf_codec.py": 3,
        "mf_processor.py": 4,
        "mf_orchestrator.py": 5,
    }
    direct_deps = {"mf_core.py": 1, "mf_transport.py": 2}
    transport_fh = direct_deps["mf_transport.py"]
    views, database = setup_multi_file_database(all_files)
    view = views[transport_fh]
    symbol_id = view.symbols.symbol_to_id.get("SENTINEL")
    assert symbol_id is not None
    ref_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=False)
    assert ref_nodes

    output = resolve_definitions_multi_file(
        view,
        database,
        request=MultiFileResolutionRequest(
            file_handle=transport_fh,
            ref_node_index=ref_nodes[0],
            include_stats=True,
            config=ForwardStitcherConfig(
                max_phases=QUERY_MAX_PHASES,
                max_total_work=QUERY_MAX_WORK,
            ),
        ),
        views=views,
    )
    assert isinstance(output, tuple)
    _, stats = output
    assert stats.files_considered == len(direct_deps)


@pytest.mark.blends_test_group("stack_unittesting")
def test_absent_dependency_does_not_load_extra_files() -> None:
    file_map = {"mf_transport.py": 2}
    transport_fh = file_map["mf_transport.py"]
    views, database = setup_multi_file_database(file_map)
    view = views[transport_fh]
    symbol_id = view.symbols.symbol_to_id.get("SENTINEL")
    assert symbol_id is not None
    ref_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=False)
    assert ref_nodes

    output = resolve_definitions_multi_file(
        view,
        database,
        request=MultiFileResolutionRequest(
            file_handle=transport_fh,
            ref_node_index=ref_nodes[0],
            include_stats=True,
            config=ForwardStitcherConfig(
                max_phases=QUERY_MAX_PHASES,
                max_total_work=QUERY_MAX_WORK,
            ),
        ),
        views=views,
    )
    assert isinstance(output, tuple)
    _, stats = output
    assert stats.files_considered == len(file_map)



import pytest

from blends.models import SyntaxGraph
from blends.stack.attributes import (
    STACK_GRAPH_IS_DEFINITION,
    STACK_GRAPH_IS_REFERENCE,
    STACK_GRAPH_KIND,
    STACK_GRAPH_SYMBOL,
)
from blends.stack.node_kinds import StackGraphNodeKind
from blends.stack.partial_path_database import PartialPathDatabase
from blends.stack.partial_path_indexer import (
    PartialPathIndexRequest,
    index_partial_paths_for_file,
)


def _build_reference_definition_graph() -> SyntaxGraph:
    graph = SyntaxGraph()
    graph.add_node("1")
    graph.update_node_attribute("1", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("1", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("1", STACK_GRAPH_IS_REFERENCE, value=True)

    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("2", STACK_GRAPH_IS_DEFINITION, value=True)

    graph.add_edge("1", "2", precedence=0)
    return graph


@pytest.mark.blends_test_group("stack_unittesting")
def test_index_partial_paths_for_file_stores_results() -> None:
    graph = _build_reference_definition_graph()
    database = PartialPathDatabase()
    request = PartialPathIndexRequest(
        file_path="test.py",
        file_handle=1,
    )

    result = index_partial_paths_for_file(graph, request, database)

    record = database.get_file(1)
    assert record is not None
    assert record.paths == result.paths



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.unary_expression import (
    build_unary_expression_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    c_ids = adj_ast(graph, args.n_id)
    expected_children_for_prefix_expression = 2

    if len(c_ids) == expected_children_for_prefix_expression:
        prefix = node_to_str(graph, c_ids[0])
        expression_id = c_ids[1]
    else:
        expression_id = c_ids[0]
        prefix = "Undefined"

    return build_unary_expression_node(args, prefix, expression_id)



from blends.models import (
    NId,
)
from blends.syntax.builders.annotation import (
    build_annotation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    annotation_id = args.ast_graph.nodes[args.n_id]["label_field_name"]
    annotation_name = node_to_str(args.ast_graph, annotation_id)
    args_id = args.ast_graph.nodes[args.n_id].get("label_field_arguments")

    return build_annotation_node(args, annotation_name, args_id)



from blends.models import (
    NId,
)
from blends.syntax.builders.symbol_lookup import (
    build_symbol_lookup_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    label_type = n_attrs.get("label_type")
    if label_type == "wildcard":
        symbol = node_to_str(graph, args.n_id)
    else:
        symbol = args.ast_graph.nodes[args.n_id].get("label_text", "Unknown")
    return build_symbol_lookup_node(args, symbol)



from __future__ import annotations

from typing import TYPE_CHECKING

from blends.stack.partial_path.partial_path import PartialPath
from blends.stack.partial_path.partial_stacks import (
    PartialScopedSymbol,
    PartialScopeStack,
    PartialSymbolStack,
)
from blends.stack.partial_path.variables import ScopeStackVariable, SymbolStackVariable

if TYPE_CHECKING:
    from blends.stack.partial_path_database import FileHandle
    from blends.stack.partial_path_db import PartialPathDb
    from blends.stack.view import StackGraphView

_BRIDGE_SYMBOL_VAR_ID = 1
_BRIDGE_SCOPE_ATTACHED_VAR_ID = 2
_BRIDGE_SCOPE_STACK_VAR_ID = 1


def _make_bridge_partial_path(
    start_node_index: int,
    pre_bound_symbol_id: int,
    post_bound_symbol_id: int,
    module_symbol_id: int,
) -> PartialPath:
    symbol_var = SymbolStackVariable(var_id=_BRIDGE_SYMBOL_VAR_ID)
    scope_attached_var = ScopeStackVariable(var_id=_BRIDGE_SCOPE_ATTACHED_VAR_ID)
    scope_stack_var = ScopeStackVariable(var_id=_BRIDGE_SCOPE_STACK_VAR_ID)

    pre_symbol_stack = PartialSymbolStack(
        symbols=(
            PartialScopedSymbol(
                symbol_id=pre_bound_symbol_id,
                scopes=PartialScopeStack.from_variable(scope_attached_var),
            ),
        ),
        variable=symbol_var,
    )
    post_symbol_stack = PartialSymbolStack(
        symbols=(
            PartialScopedSymbol(symbol_id=module_symbol_id, scopes=None),
            PartialScopedSymbol(
                symbol_id=post_bound_symbol_id,
                scopes=PartialScopeStack.from_variable(scope_attached_var),
            ),
        ),
        variable=symbol_var,
    )
    pre_scope_stack = PartialScopeStack.from_variable(scope_stack_var)
    post_scope_stack = PartialScopeStack.from_variable(scope_attached_var)

    return PartialPath(
        start_node_index=start_node_index,
        end_node_index=0,
        symbol_stack_precondition=pre_symbol_stack,
        symbol_stack_postcondition=post_symbol_stack,
        scope_stack_precondition=pre_scope_stack,
        scope_stack_postcondition=post_scope_stack,
        edges=(),
    )


def index_bridge_paths_for_file(
    view: StackGraphView,
    *,
    file_handle: FileHandle,
    db: PartialPathDb,
) -> None:
    for node_index, is_bridge in enumerate(view.node_is_import_binding):
        if not is_bridge:
            continue
        module_stem = view.node_import_module_stem[node_index]
        if not module_stem:
            continue
        module_symbol_id = db.symbols.symbol_to_id.get(module_stem)
        if module_symbol_id is None:
            continue
        pre_bound_symbol_id = view.node_symbol_id[node_index]
        if pre_bound_symbol_id is None:
            continue
        import_source_symbol = view.node_import_source_symbol[node_index]
        source_symbol_str = (
            import_source_symbol
            if import_source_symbol is not None
            else view.symbols.id_to_symbol[pre_bound_symbol_id]
        )
        post_bound_symbol_id = db.symbols.symbol_to_id.get(source_symbol_str)
        if post_bound_symbol_id is None:
            post_bound_symbol_id = db.symbols.intern(source_symbol_str)
        bridge = _make_bridge_partial_path(
            node_index, pre_bound_symbol_id, post_bound_symbol_id, module_symbol_id
        )
        db.add_bridge_path(file_handle, bridge)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.binary_operation import (
    build_binary_operation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    child_ids = adj_ast(graph, args.n_id)
    left_id = None
    operator = ""
    right_id = None
    min_children_for_binary_operation = 2
    if len(child_ids) > min_children_for_binary_operation:
        left_id = child_ids[0]
        operator_id = child_ids[1]
        right_id = child_ids[2]
        operator = graph.nodes[operator_id]["label_text"]
    return build_binary_operation_node(args, operator, left_id, right_id)



from blends.models import (
    NId,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    n_attrs = args.ast_graph.nodes[args.n_id]
    var_name = node_to_str(args.ast_graph, n_attrs["label_field_name"])

    return build_variable_declaration_node(args, var_name, None, None)



from __future__ import annotations

from mf_processor import Processor, ProcessorConfig, ProcessingResult, ProcessorError
from mf_transport import ConnectionPool, Request, Response, send_with_retry
from mf_codec import CodecError
from mf_core import MAX_RETRIES, SENTINEL, ConnectionConfig, RetryPolicy
from mf_core import CHUNK_SIZE as BUFFER_SIZE

OrchestratorError = ProcessorError


class OrchestratorStats:
    total_requests: int = 0
    successful_requests: int = 0
    failed_requests: int = 0
    total_frames: int = 0
    sentinel_hits: int = 0
    codec_errors: int = 0

    def success_rate(self: "OrchestratorStats") -> int:
        if self.total_requests == 0:
            return 100
        return self.successful_requests * 100 // self.total_requests

    def avg_frames(self: "OrchestratorStats") -> int:
        if self.successful_requests == 0:
            return 0
        return self.total_frames // self.successful_requests

    def is_healthy(self: "OrchestratorStats") -> bool:
        rate: int = self.success_rate()
        return rate >= 95

    def reset(self: "OrchestratorStats") -> None:
        self.total_requests = 0
        self.successful_requests = 0
        self.failed_requests = 0
        self.total_frames = 0
        self.sentinel_hits = 0
        self.codec_errors = 0

class Orchestrator:
    processor: Processor = Processor()
    pool: ConnectionPool = ConnectionPool()
    stats: OrchestratorStats = OrchestratorStats()
    config: ProcessorConfig = ProcessorConfig()
    max_retries: int = MAX_RETRIES
    buffer_size: int = BUFFER_SIZE

    def initialize(self: "Orchestrator", conn_config: ConnectionConfig) -> None:
        self.pool.config = conn_config
        self.config.pool = self.pool
        self.processor.config = self.config

    def handle(self: "Orchestrator", payload_size: int) -> bool:
        self.stats.total_requests = self.stats.total_requests + 1
        result: ProcessingResult = self.processor.run(payload_size)
        if result.is_ok():
            self.stats.successful_requests = self.stats.successful_requests + 1
            self.stats.total_frames = self.stats.total_frames + result.frames_encoded
            return True
        self.stats.failed_requests = self.stats.failed_requests + 1
        return False

    def handle_batch(self: "Orchestrator", sizes: list) -> int:
        success: int = self.processor.run_batch(sizes)
        self.stats.total_requests = self.stats.total_requests + len(sizes)
        self.stats.successful_requests = self.stats.successful_requests + success
        self.stats.failed_requests = self.stats.failed_requests + len(sizes) - success
        return success

    def get_health(self: "Orchestrator") -> int:
        return self.stats.success_rate()

    def direct_send(self: "Orchestrator", payload_size: int) -> Response:
        req: Request = Request()
        req.payload_size = payload_size
        retry: RetryPolicy = RetryPolicy()
        retry.max_attempts = self.max_retries
        req.retry = retry
        resp: Response = send_with_retry(self.pool, req)
        if resp.status == SENTINEL:
            self.stats.sentinel_hits = self.stats.sentinel_hits + 1
        return resp

    def record_codec_error(self: "Orchestrator", err: CodecError) -> None:
        if err.is_fatal():
            self.stats.codec_errors = self.stats.codec_errors + 1


def build_orchestrator(host: str, port: int) -> Orchestrator:
    orch: Orchestrator = Orchestrator()
    conn: ConnectionConfig = ConnectionConfig()
    conn.host = host
    conn.port = port
    conn.max_connections = MAX_RETRIES * 10
    policy: RetryPolicy = RetryPolicy()
    policy.max_attempts = MAX_RETRIES
    orch.config.retry = policy
    orch.initialize(conn)
    return orch

GLOBAL_ORCHESTRATOR: Orchestrator = Orchestrator()
GLOBAL_POOL: ConnectionPool = ConnectionPool()

SENTINEL_STATUS: int = SENTINEL
MAX_RETRY_COUNT: int = MAX_RETRIES
BUFFER: int = BUFFER_SIZE
ORCH_ERROR: OrchestratorError = OrchestratorError()



from collections.abc import (
    Iterator,
)

from blends.query import (
    adj_cfg,
)
from blends.symbolic_evaluation.context.search.model import (
    SearchArgs,
    SearchResult,
)

VAR_MAPPINGS = {
    "VariableDeclaration": "variable",
    "MethodDeclaration": "name",
    "Import": "expression",
    "ModuleImport": "expression",
}


def search(args: SearchArgs) -> Iterator[SearchResult]:
    for c_id in adj_cfg(args.graph, args.n_id):
        n_attrs = args.graph.nodes[c_id]
        if n_attrs["label_type"] in VAR_MAPPINGS and (
            n_attrs.get(VAR_MAPPINGS[n_attrs["label_type"]]) == args.symbol
            or n_attrs.get("label_alias") == args.symbol
        ):
            yield True, c_id
            break



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.assignment import (
    build_assignment_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    as_attrs = args.ast_graph.nodes[args.n_id]
    var_id = as_attrs["label_field_target"]
    if (
        graph.nodes[var_id]["label_type"] == "directly_assignable_expression"
        and (childs := adj_ast(graph, var_id))
        and len(childs) == 1
    ):
        var_id = childs[0]
    val_id = as_attrs.get("label_field_result")
    operator_id = as_attrs.get("label_field_operator")
    operator = node_to_str(graph, operator_id) if operator_id else None
    if not val_id:
        val_id = adj_ast(graph, args.n_id)[-1]
    return build_assignment_node(args, var_id, val_id, operator)



from blends.ctx import ctx
from blends.models import (
    NId,
)
from blends.stack.edges import (
    Edge,
    add_edge,
)
from blends.stack.node_helpers import (
    pop_scoped_symbol_node_attributes,
)
from blends.syntax.builders.utils import (
    add_pattern_binding_nodes,
    bound_identifier_symbol,
    get_next_binding_precedence,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_for_statement_node(
    args: SyntaxGraphArgs,
    initializer_node: NId | None,
    condition_node: NId | None,
    update_node: NId | None,
    body_node: NId,
) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        label_type="ForStatement",
        block_id=body_node,
    )

    args.syntax_graph.add_edge(
        args.n_id,
        args.generic(args.fork_n_id(body_node)),
        label_ast="AST",
    )

    if initializer_node:
        args.syntax_graph.update_node_attribute(args.n_id, "initializer_id", initializer_node)
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(initializer_node)),
            label_ast="AST",
        )

    if condition_node:
        args.syntax_graph.update_node_attribute(args.n_id, "condition_id", condition_node)
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(condition_node)),
            label_ast="AST",
        )

    if update_node:
        args.syntax_graph.update_node_attribute(args.n_id, "update_id", update_node)
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(update_node)),
            label_ast="AST",
        )

    if ctx.has_feature_flag("StackGraph") and initializer_node is not None:
        if (symbol := bound_identifier_symbol(args, var_id=initializer_node)) is not None:
            precedence = get_next_binding_precedence(args)
            args.syntax_graph.update_node(
                args.n_id,
                pop_scoped_symbol_node_attributes(
                    symbol=symbol,
                    precedence=precedence,
                ),
            )
            scope_stack = args.metadata.setdefault("scope_stack", [])
            if scope_stack and (parent_scope := scope_stack[-1]):
                add_edge(
                    args.syntax_graph,
                    Edge(source=parent_scope, sink=args.n_id, precedence=precedence),
                )
        else:
            add_pattern_binding_nodes(args, var_id=initializer_node)

    return args.n_id



from collections.abc import (
    Callable,
    Iterable,
    Iterator,
)
from itertools import (
    chain,
)

from blends.models import (
    Graph,
    NId,
)

NAttrs = dict[str, str]


def has_labels(n_attrs: NAttrs, **expected_attrs: str) -> bool:
    return all(
        n_attrs.get(expected_attr) == expected_attr_value
        for expected_attr, expected_attr_value in expected_attrs.items()
    )


def predicate_has_labels(**expected_attrs: str) -> Callable[[NAttrs], bool]:
    def predicate(n_attrs: NAttrs) -> bool:
        return has_labels(n_attrs, **expected_attrs)

    return predicate


def filter_nodes(
    graph: Graph,
    nodes: Iterable[NId],
    predicate: Callable[[NAttrs], bool],
) -> tuple[NId, ...]:
    result: tuple[NId, ...] = tuple(n_id for n_id in nodes if predicate(graph.nodes[n_id]))

    return result


def matching_nodes(
    graph: Graph, label_type: str | None = None, **expected_attrs: str
) -> tuple[NId, ...]:
    if label_type:
        candidates = graph.nodes_by_type.get(label_type, [])
        if expected_attrs:
            return filter_nodes(graph, candidates, predicate_has_labels(**expected_attrs))
        return tuple(candidates)
    return filter_nodes(graph, graph.nodes, predicate_has_labels(**expected_attrs))


def _edge_matches(
    graph: Graph,
    source: NId,
    target: NId,
    edge_keys: set[str],
    *,
    strict: bool,
    **edge_attrs: str,
) -> bool:
    if not has_labels(graph[source][target], **edge_attrs):
        return False
    if strict:
        graph_edge_keys = set(graph[source][target].keys())
        difference = graph_edge_keys.difference(edge_keys)
        return not difference or difference == {"label_index"}
    return True


def adj_lazy(
    graph: Graph,
    n_id: NId,
    depth: int = 1,
    *,
    strict: bool = False,
    _processed_n_ids: set[NId] | None = None,
    **edge_attrs: str,
) -> Iterator[NId]:
    """Return adjacent nodes to `n_id`, following just edges with given attrs.

    - Parameter `depth` may be -1 to indicate infinite depth.
    - Parameter `strict` indicates that the edges must have only the indicated
      attributes

    - Search is done breadth first.
    - Nodes are returned ordered ascending by index on each level.

    This function must be used instead of graph.adj, because graph.adj
    becomes unstable (unordered) after mutating the graph, also this allow
    following just edges matching `edge_attrs`.
    """
    processed_n_ids: set[NId] = _processed_n_ids or set()
    if depth == 0 or n_id in processed_n_ids:
        return

    processed_n_ids.add(n_id)

    childs: list[NId] = sorted(graph.adj[n_id], key=int)
    edge_keys = set(edge_attrs.keys())
    matching_childs = [
        c_id
        for c_id in childs
        if _edge_matches(graph, n_id, c_id, edge_keys, strict=strict, **edge_attrs)
    ]

    yield from matching_childs

    if depth in {0, 1}:
        return

    for c_id in matching_childs:
        yield from adj_lazy(
            graph,
            c_id,
            depth=depth - 1,
            strict=strict,
            _processed_n_ids=processed_n_ids,
            **edge_attrs,
        )


def adj(
    graph: Graph,
    n_id: NId,
    depth: int = 1,
    *,
    strict: bool = False,
    _processed_n_ids: set[NId] | None = None,
    **edge_attrs: str,
) -> tuple[NId, ...]:
    return tuple(
        adj_lazy(
            graph,
            n_id,
            depth=depth,
            strict=strict,
            _processed_n_ids=_processed_n_ids,
            **edge_attrs,
        ),
    )


def adj_ast(
    graph: Graph,
    n_id: NId,
    depth: int = 1,
    *,
    strict: bool = False,
    **n_attrs: str,
) -> tuple[NId, ...]:
    return tuple(
        c_id
        for c_id in adj(graph, n_id, depth, strict=strict, label_ast="AST")
        if has_labels(graph.nodes[c_id], **n_attrs)
    )


def adj_cfg(
    graph: Graph,
    n_id: NId,
    depth: int = 1,
    *,
    strict: bool = False,
    **n_attrs: str,
) -> tuple[NId, ...]:
    return tuple(
        c_id
        for c_id in adj(graph, n_id, depth, strict=strict, label_cfg="CFG")
        if has_labels(graph.nodes[c_id], **n_attrs)
    )


def adj_cfg_lazy(
    graph: Graph,
    n_id: NId,
    depth: int = 1,
    *,
    strict: bool = False,
    **n_attrs: str,
) -> Iterator[NId]:
    yield from adj_lazy(
        graph,
        n_id,
        depth=depth,
        strict=strict,
        _processed_n_ids=set(),
        label_cfg="CFG",
        **n_attrs,
    )


def pred_lazy(
    graph: Graph,
    n_id: NId,
    depth: int = 1,
    _processed_n_ids: set[NId] | None = None,
    **edge_attrs: str,
) -> Iterator[NId]:
    """Obtain parent nodes.

    Same as `adj` but follow edges in the opposite direction
    """
    processed_n_ids: set[NId] = _processed_n_ids or set()
    if depth == 0 or n_id in processed_n_ids:
        return

    processed_n_ids.add(n_id)

    p_ids: list[NId] = sorted(graph.pred[n_id], key=int)
    matching_parents = [p_id for p_id in p_ids if has_labels(graph[p_id][n_id], **edge_attrs)]

    yield from matching_parents

    if depth < 0 or depth > 1:
        for p_id in matching_parents:
            yield from pred_lazy(
                graph,
                p_id,
                depth=depth - 1,
                _processed_n_ids=processed_n_ids,
                **edge_attrs,
            )


def pred(
    graph: Graph,
    n_id: NId,
    depth: int = 1,
    _processed_n_ids: set[NId] | None = None,
    **edge_attrs: str,
) -> tuple[NId, ...]:
    return tuple(
        pred_lazy(
            graph,
            n_id,
            depth,
            _processed_n_ids=_processed_n_ids,
            **edge_attrs,
        ),
    )


def pred_ast(
    graph: Graph,
    n_id: NId,
    depth: int = 1,
    **edge_attrs: str,
) -> tuple[NId, ...]:
    return tuple(pred_ast_lazy(graph, n_id, depth, **edge_attrs))


def pred_ast_lazy(
    graph: Graph,
    n_id: NId,
    depth: int = 1,
    **edge_attrs: str,
) -> Iterator[NId]:
    yield from pred_lazy(
        graph,
        n_id,
        depth,
        _processed_n_ids=set(),
        label_ast="AST",
        **edge_attrs,
    )


def pred_cfg(
    graph: Graph,
    n_id: NId,
    depth: int = 1,
    **edge_attrs: str,
) -> tuple[NId, ...]:
    return tuple(pred_cfg_lazy(graph, n_id, depth, **edge_attrs))


def pred_cfg_lazy(
    graph: Graph,
    n_id: NId,
    depth: int = 1,
    **edge_attrs: str,
) -> Iterator[NId]:
    yield from pred_lazy(
        graph,
        n_id,
        depth,
        _processed_n_ids=set(),
        label_cfg="CFG",
        **edge_attrs,
    )


def match_ast(
    graph: Graph,
    n_id: NId,
    *label_type: str,
    depth: int = 1,
) -> dict[str, NId | None]:
    index: int = 0
    nodes: dict[str, NId | None] = dict.fromkeys(label_type)

    for c_id in adj_ast(graph, n_id, depth=depth):
        c_type = graph.nodes[c_id]["label_type"]
        if c_type in nodes and nodes[c_type] is None:
            nodes[c_type] = c_id
        else:
            nodes[f"__{index}__"] = c_id
            index += 1

    return nodes


def match_ast_d(graph: Graph, n_id: NId, label_type: str, depth: int = 1) -> NId | None:
    return match_ast(graph, n_id, label_type, depth=depth)[label_type]


def match_ast_group(
    graph: Graph,
    n_id: NId,
    *label_type: str,
    depth: int = 1,
) -> dict[str, list[NId]]:
    index: int = 0
    nodes: dict[str, list[NId]] = {label: [] for label in label_type}

    for c_id in adj_ast(graph, n_id, depth=depth):
        c_type = graph.nodes[c_id]["label_type"]
        if c_type in nodes:
            nodes[c_type].append(c_id)
        else:
            nodes[f"__{index}__"] = [
                c_id,
            ]
            index += 1

    return nodes


def match_ast_group_d(graph: Graph, n_id: NId, label_type: str, depth: int = 1) -> list[NId]:
    return match_ast_group(graph, n_id, label_type, depth=depth)[label_type]


def is_connected_to_cfg(graph: Graph, n_id: NId) -> bool:
    return bool(adj_cfg(graph, n_id) or pred_cfg(graph, n_id))


def lookup_first_cfg_parent(graph: Graph, n_id: NId) -> NId:
    # Lookup first parent who is connected to the CFG
    for p_id in chain([n_id], pred_ast_lazy(graph, n_id, depth=-1)):
        if is_connected_to_cfg(graph, p_id):
            return p_id
    # Base case, pass through
    return n_id


def _get_subgraph(
    graph: Graph,
    node_n_id_predicate: Callable[[NId], bool] = lambda _: True,
    edge_n_attrs_predicate: Callable[[NAttrs], bool] = lambda _: True,
) -> Graph:
    copy: Graph = Graph()

    for n_a_id, n_b_id in graph.edges:
        edge_attrs = graph[n_a_id][n_b_id].copy()
        n_a_attrs = graph.nodes[n_a_id].copy()
        n_b_attrs = graph.nodes[n_b_id].copy()

        if (
            edge_n_attrs_predicate(edge_attrs)
            and node_n_id_predicate(n_a_id)
            and node_n_id_predicate(n_b_id)
        ):
            copy.add_node(n_a_id, **n_a_attrs)
            copy.add_node(n_b_id, **n_b_attrs)
            copy.add_edge(n_a_id, n_b_id, **edge_attrs)

    return copy


def copy_ast(graph: Graph) -> Graph:
    return _get_subgraph(
        graph=graph,
        edge_n_attrs_predicate=predicate_has_labels(label_ast="AST"),
    )


def search_pred_until_type(
    graph: Graph,
    n_id: NId,
    targets: set[str],
    last_child: NId | None = None,
) -> tuple[NId, NId] | None:
    if not last_child:
        last_child = n_id
    if pred_c := pred_ast(graph, n_id):
        if graph.nodes[pred_c[0]]["label_type"] in targets:
            return (pred_c[0], last_child)
        return search_pred_until_type(graph, pred_c[0], targets, pred_c[0])
    return None


def get_nodes_by_path(
    graph: Graph,
    n_id: NId,
    *label_type_path: str,
) -> set[NId]:
    if not label_type_path:
        return set()
    if len(label_type_path) == 1:
        return set(adj_ast(graph, n_id, label_type=label_type_path[0]))

    result: set[NId] = set()
    for node in adj_ast(graph, n_id, label_type=label_type_path[0]):
        result.update(get_nodes_by_path(graph, node, *label_type_path[1:]))
    return result


def get_node_by_path(graph: Graph, n_id: NId, *label_type_path: str) -> NId | None:
    return next(iter(get_nodes_by_path(graph, n_id, *label_type_path)), None)


def nodes_by_type(graph: Graph) -> dict[str, list[NId]]:
    return graph.nodes_by_type



from blends.models import (
    NId,
)
from blends.syntax.builders.class_decl import (
    build_class_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]

    name_str = ""
    if name_id := n_attrs.get("label_field_name"):
        name_str = node_to_str(graph, name_id)

    elif val_id := n_attrs.get("label_field_value"):
        name_str = "<<" + node_to_str(graph, val_id)

    block_id = n_attrs.get("label_field_body")

    superclass_name_str = None
    if superclass_id := n_attrs.get("label_field_superclass"):
        superclass_name_str = node_to_str(graph, superclass_id)[1:]

    return build_class_node(args, name_str, block_id, None, superclass_name_str)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.return_statement import (
    build_return_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    match = match_ast(args.ast_graph, args.n_id, "return", ";")
    expected_match_count_for_return_with_value = 3
    if len(match) == expected_match_count_for_return_with_value:
        return build_return_node(args, value_id=match["__0__"])
    return build_return_node(args, None)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.method_invocation import (
    DirectChildren,
    build_method_invocation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    expr_id = adj_ast(graph, args.n_id)[0]
    expr = node_to_str(graph, expr_id)

    arguments_id = match_ast_d(graph, args.n_id, "call_suffix")

    if (
        arguments_id
        and (childs := adj_ast(graph, arguments_id))
        and len(childs) == 1
        and graph.nodes[childs[0]]["label_type"] == "value_arguments"
    ):
        arguments_id = childs[0]

    children: DirectChildren = {"expression_id": expr_id}
    if arguments_id:
        children["arguments_id"] = arguments_id

    return build_method_invocation_node(args, expr, children)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.switch_section import (
    build_switch_section_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    filtered_labels = {
        "\n",
        "\r\n",
        "case",
        ":",
        "default",
    }
    childs = adj_ast(graph, args.n_id)
    value_id = graph.nodes[args.n_id].get("label_field_value") or graph.nodes[args.n_id].get(
        "label_field_type",
    )
    if not value_id:
        value_id = childs[0]
    case_expr = graph.nodes[value_id].get("label_text") or node_to_str(graph, value_id)

    execution_ids = [
        _id
        for _id in childs
        if graph.nodes[_id]["label_type"] not in filtered_labels and _id != value_id
    ]

    return build_switch_section_node(args, case_expr, iter(execution_ids))



import pytest

from blends.ctx import ctx
from blends.models import NId, SyntaxGraph
from blends.symbolic_evaluation import evaluate
from blends.symbolic_evaluation.dispatcher import generic
from blends.symbolic_evaluation.models import (
    MissingSymbolicEvalError,
    SymbolicEvalArgs,
    SymbolicEvaluation,
)
from blends.symbolic_evaluation.utils import (
    get_backward_paths,
    get_current_class,
    get_forward_paths,
    get_lookup_path,
    get_object_identifiers,
    iter_backward_paths,
    iter_forward_paths,
)


def add_ast_edge(graph: SyntaxGraph, src: str, dst: str) -> None:
    graph.add_edge(src, dst, label_ast="AST")


def add_cfg_edge(graph: SyntaxGraph, src: str, dst: str) -> None:
    graph.add_edge(src, dst, label_cfg="CFG")


def base_args(graph: SyntaxGraph, n_id: NId, path: list[str]) -> SymbolicEvalArgs:
    return SymbolicEvalArgs(
        evaluation={},
        graph=graph,
        path=path,
        n_id=n_id,
        triggers=set(),
        graph_db=None,
        method_evaluators=None,
    )


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_generic_missing_evaluator_raises() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="UnknownType")
    args = base_args(graph, "1", ["1"])

    with pytest.raises(MissingSymbolicEvalError):
        generic(args)


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_evaluate_returns_none_on_missing_evaluator() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="UnknownType")

    result = evaluate(graph, ["1"], "1")

    assert result is None


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_iter_backward_paths_single_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="ExpressionStatement")

    paths = list(iter_backward_paths(graph, "1"))

    assert paths == [["1"]]


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_iter_backward_paths_linear_chain() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="ExpressionStatement")
    graph.add_node("2", label_type="ExpressionStatement")
    graph.add_node("3", label_type="ExpressionStatement")
    add_cfg_edge(graph, "1", "2")
    add_cfg_edge(graph, "2", "3")

    paths = list(iter_backward_paths(graph, "3"))

    assert paths == [["3", "2", "1"]]


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_iter_backward_paths_branching_parents() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="ExpressionStatement")
    graph.add_node("2", label_type="ExpressionStatement")
    graph.add_node("3", label_type="ExpressionStatement")
    add_cfg_edge(graph, "1", "3")
    add_cfg_edge(graph, "2", "3")

    paths = list(iter_backward_paths(graph, "3"))

    assert paths == [["3", "1"], ["3", "2"]]


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_get_backward_paths_uses_cfg_parent_from_ast() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="ExpressionStatement")
    graph.add_node("2", label_type="ExpressionStatement")
    graph.add_node("3", label_type="Literal")
    add_ast_edge(graph, "2", "3")
    add_cfg_edge(graph, "1", "2")

    paths = list(get_backward_paths(graph, "3"))

    assert paths == [["2", "1"]]


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_get_backward_paths_respects_recursion_limit() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="ExpressionStatement")
    graph.add_node("2", label_type="ExpressionStatement")
    graph.add_node("3", label_type="ExpressionStatement")
    add_cfg_edge(graph, "1", "2")
    add_cfg_edge(graph, "2", "3")
    original_limit = ctx.recursion_limit
    ctx.recursion_limit = 1
    try:
        paths = list(get_backward_paths(graph, "3"))
    finally:
        ctx.recursion_limit = original_limit

    assert paths == []


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_iter_forward_paths_single_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="ExpressionStatement")

    paths = list(iter_forward_paths(graph, "1"))

    assert paths == [["1"]]


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_iter_forward_paths_linear_chain() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="ExpressionStatement")
    graph.add_node("2", label_type="ExpressionStatement")
    graph.add_node("3", label_type="ExpressionStatement")
    add_cfg_edge(graph, "1", "2")
    add_cfg_edge(graph, "2", "3")

    paths = list(iter_forward_paths(graph, "1"))

    assert paths == [["1", "2", "3"]]


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_iter_forward_paths_branching_children() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="ExpressionStatement")
    graph.add_node("2", label_type="ExpressionStatement")
    graph.add_node("3", label_type="ExpressionStatement")
    add_cfg_edge(graph, "1", "2")
    add_cfg_edge(graph, "1", "3")

    paths = list(iter_forward_paths(graph, "1"))

    assert paths == [["1", "2"], ["1", "3"]]


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_get_forward_paths_uses_cfg_parent_from_ast() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="ExpressionStatement")
    graph.add_node("2", label_type="ExpressionStatement")
    graph.add_node("3", label_type="Literal")
    add_ast_edge(graph, "2", "3")
    add_cfg_edge(graph, "1", "2")

    paths = list(get_forward_paths(graph, "3"))

    assert paths == [["2"]]


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_get_lookup_path_returns_tail_after_cfg_parent() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="SymbolLookup", symbol="x")
    graph.add_node("2", label_type="ExpressionStatement")
    add_cfg_edge(graph, "1", "2")
    path = ["1", "9", "10"]

    result = get_lookup_path(graph, path, "1")

    assert result == ["9", "10"]


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_get_lookup_path_returns_empty_when_parent_last() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="SymbolLookup", symbol="x")
    graph.add_node("2", label_type="ExpressionStatement")
    add_cfg_edge(graph, "1", "2")
    path = ["9", "1"]

    result = get_lookup_path(graph, path, "1")

    assert result == []


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_get_object_identifiers_returns_variable_names() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="VariableDeclaration", variable="client")
    graph.add_node("2", label_type="ObjectCreation", name="HttpClient")
    add_ast_edge(graph, "1", "2")

    result = get_object_identifiers(graph, {"HttpClient"})

    assert result == ["client"]


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_get_object_identifiers_ignores_unlisted_name() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="VariableDeclaration", variable="client")
    graph.add_node("2", label_type="ObjectCreation", name="HttpClient")
    add_ast_edge(graph, "1", "2")

    result = get_object_identifiers(graph, {"OtherClient"})

    assert result == []


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_get_object_identifiers_ignores_non_variable_declaration_parent() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="Literal")
    graph.add_node("2", label_type="ObjectCreation", name="HttpClient")
    add_ast_edge(graph, "1", "2")

    result = get_object_identifiers(graph, {"HttpClient"})

    assert result == []


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_get_object_identifiers_ignores_missing_variable_name() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="VariableDeclaration")
    graph.add_node("2", label_type="ObjectCreation", name="HttpClient")
    add_ast_edge(graph, "1", "2")

    result = get_object_identifiers(graph, {"HttpClient"})

    assert result == []


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_get_current_class_returns_class_parent() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="Class")
    graph.add_node("2", label_type="MethodDeclaration")
    graph.add_node("3", label_type="Literal")
    add_ast_edge(graph, "1", "2")
    add_ast_edge(graph, "2", "3")

    result = get_current_class(graph, "3")

    assert result == "1"


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_get_current_class_returns_self_when_no_class_parent() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="MethodDeclaration")
    graph.add_node("2", label_type="Literal")
    add_ast_edge(graph, "1", "2")

    result = get_current_class(graph, "2")

    assert result == "2"


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_symbolic_eval_args_fork_n_id_preserves_state() -> None:
    graph = SyntaxGraph()
    evaluation = {"1": True}
    triggers = {"source"}
    args = SymbolicEvalArgs(
        evaluation=evaluation,
        graph=graph,
        path=["1", "2"],
        n_id="1",
        triggers=triggers,
        graph_db=None,
        method_evaluators=None,
    )

    forked = args.fork_n_id("2")

    assert forked.n_id == "2"
    assert forked.path is args.path
    assert forked.graph is args.graph
    assert forked.evaluation is args.evaluation
    assert forked.triggers is args.triggers


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_symbolic_eval_args_fork_n_id_overrides_graph_and_path() -> None:
    graph = SyntaxGraph()
    next_graph = SyntaxGraph()
    args = SymbolicEvalArgs(
        evaluation={},
        graph=graph,
        path=["1"],
        n_id="1",
        triggers=set(),
        graph_db=None,
        method_evaluators=None,
    )

    forked = args.fork_n_id("2", path=["2"], graph=next_graph)

    assert forked.n_id == "2"
    assert forked.path == ["2"]


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_symbolic_eval_args_fork_n_id_empty_path_uses_original() -> None:
    graph = SyntaxGraph()
    args = SymbolicEvalArgs(
        evaluation={},
        graph=graph,
        path=["1"],
        n_id="1",
        triggers=set(),
        graph_db=None,
        method_evaluators=None,
    )

    forked = args.fork_n_id("2", path=[])

    assert forked.path is args.path


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_symbolic_eval_args_fork_updates_selected_fields() -> None:
    graph = SyntaxGraph()
    args = SymbolicEvalArgs(
        evaluation={"1": False},
        graph=graph,
        path=["1"],
        n_id="1",
        triggers={"t1"},
        graph_db=None,
        method_evaluators=None,
    )

    forked = args.fork(n_id="2")

    assert forked.n_id == "2"
    assert forked.graph is args.graph
    assert forked.path is args.path
    assert forked.evaluation is args.evaluation
    assert forked.triggers is args.triggers


@pytest.mark.blends_test_group("symbolic_evaluation_unittesting")
def test_symbolic_evaluation_fields() -> None:
    evaluation = SymbolicEvaluation(danger=True, triggers={"a"})

    assert evaluation.danger is True
    assert evaluation.triggers == {"a"}



from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_export_statement_node(
    args: SyntaxGraphArgs,
    expression: str | None,
    exported_block: NId | None,
) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        label_type="Export",
    )

    if expression:
        args.syntax_graph.update_node_attribute(args.n_id, "expression", expression)

    if exported_block:
        args.syntax_graph.update_node_attribute(args.n_id, "declaration_id", exported_block)
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(exported_block)),
            label_ast="AST",
        )

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    get_node_by_path,
)
from blends.syntax.builders.binary_operation import (
    build_binary_operation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    field_left = get_node_by_path(
        graph,
        args.n_id,
        "type_identifier",
    )

    return build_binary_operation_node(
        args,
        "isInstanceOf",
        field_left,
        None,
    )



from blends.models import (
    NId,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    var_id = graph.nodes[args.n_id]["label_field_name"]
    var_name = node_to_str(graph, var_id)
    var_name = var_name[1:-1] if var_name.startswith("{") else var_name
    value_id = graph.nodes[args.n_id].get("label_field_value")
    return build_variable_declaration_node(args, var_name, None, value_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.parameter_list import (
    build_parameter_list_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    valid_parameters = {"identifier", "class_parameter", "binding", "parameter"}
    graph = args.ast_graph
    c_ids = adj_ast(graph, args.n_id)
    return build_parameter_list_node(
        args,
        c_ids=(_id for _id in c_ids if graph.nodes[_id]["label_type"] in valid_parameters),
    )



from blends.models import (
    SUPPORTED_MULTIFILE,
    NId,
    decide_language,
)
from blends.query import (
    adj_ast,
)
from blends.symbolic_evaluation.context.method import (
    solve_invocation,
)
from blends.symbolic_evaluation.dispatcher import (
    generic,
)
from blends.symbolic_evaluation.models import (
    SymbolicEvalArgs,
    SymbolicEvaluation,
)
from blends.symbolic_evaluation.multifile.method_invocation import (
    evaluate_method_invocation,
)

MODIFYING_METHODS: set[str] = {"add", "push", "put", "get"}


def evaluate_method_expression(
    args: SymbolicEvalArgs,
    expr_id: NId,
) -> bool:
    if args.graph.nodes[expr_id].get("symbol") in MODIFYING_METHODS:
        return False

    if md_id := solve_invocation(args.graph, args.path, expr_id):
        # Avoid infinite propagation in recursive method invocations
        if expr_id in adj_ast(args.graph, md_id, depth=-1):
            return False
        fork_id = args.graph.nodes[md_id].get("block_id") or md_id
        d_expression = generic(args.fork_n_id(fork_id)).danger
    else:
        d_expression = generic(args.fork_n_id(expr_id)).danger

    return d_expression


def evaluate(args: SymbolicEvalArgs) -> SymbolicEvaluation:
    args.evaluation[args.n_id] = False
    n_attrs = args.graph.nodes[args.n_id]

    d_method = False
    propagate_danger = True
    if (
        (metadata_node := args.graph.nodes.get(NId(0)))
        and (lang := decide_language(path=metadata_node["path"]))
        and lang in SUPPORTED_MULTIFILE
    ):
        d_method, propagate_danger = evaluate_method_invocation(args, lang)

    propagation_result = False
    if propagate_danger:
        d_expression = False
        if expr_id := n_attrs.get("expression_id"):
            d_expression = evaluate_method_expression(args, expr_id)

        d_arguments = False
        if al_id := args.graph.nodes[args.n_id].get("arguments_id"):
            d_arguments = generic(args.fork_n_id(al_id)).danger

        d_object = False
        if obj_id := args.graph.nodes[args.n_id].get("object_id"):
            d_object = generic(args.fork_n_id(obj_id)).danger

        propagation_result = d_expression or d_arguments or d_object

    args.evaluation[args.n_id] = propagation_result or d_method
    if args.method_evaluators and (
        method_evaluator := args.method_evaluators.get("method_invocation")
    ):
        args.evaluation[args.n_id] = method_evaluator(args).danger

    return SymbolicEvaluation(args.evaluation[args.n_id], args.triggers)



from pathlib import Path

from blends.ast.build_graph import get_ast_graph_from_content, get_ast_graph_from_path
from blends.content.content import get_content_by_path, get_language_by_path
from blends.models import GraphPair
from blends.syntax.build_graph import get_syntax_graph, get_syntax_graph_from_path
from blends.tree.parse import get_tree


def get_graphs_from_path(
    path: Path, *, with_cfg: bool = True, with_metadata: bool = False
) -> GraphPair:
    content = get_content_by_path(path)
    if content is None:
        return GraphPair(ast_graph=None, syntax_graph=None)

    ast_graph = get_ast_graph_from_content(content)
    if ast_graph is None:
        return GraphPair(ast_graph=None, syntax_graph=None)

    syntax_graph = get_syntax_graph(
        ast_graph, content, with_cfg=with_cfg, with_metadata=with_metadata
    )
    if syntax_graph is None:
        return GraphPair(ast_graph=ast_graph, syntax_graph=None)

    return GraphPair(ast_graph=ast_graph, syntax_graph=syntax_graph)


__all__ = [
    "get_ast_graph_from_content",
    "get_ast_graph_from_path",
    "get_content_by_path",
    "get_graphs_from_path",
    "get_language_by_path",
    "get_syntax_graph",
    "get_syntax_graph_from_path",
    "get_tree",
]



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.expression_statement import (
    build_expression_statement_node,
)
from blends.syntax.builders.symbol_lookup import (
    build_symbol_lookup_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    c_ids = [
        child
        for child in adj_ast(graph, args.n_id)
        if graph.nodes[child]["label_type"] not in {"$", "{", "}"}
    ]

    if len(c_ids) == 1 and graph.nodes[c_ids[0]]["label_type"] == "identifier_dollar_escaped":
        symbol = graph.nodes[c_ids[0]].get("label_text", "")
        return build_symbol_lookup_node(args, symbol)

    return build_expression_statement_node(args, iter(c_ids))



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
    match_ast_d,
)
from blends.syntax.builders.binary_operation import (
    build_binary_operation_node,
)
from blends.syntax.builders.expression_statement import (
    build_expression_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    c_ids = adj_ast(graph, args.n_id)
    invalid_childs = {";", "type_cast", "(", ")"}
    expected_children_for_type_cast = 2

    if (
        len(c_ids) == expected_children_for_type_cast
        and (type_cast := match_ast_d(graph, args.n_id, "type_cast"))
        and (typed_assignment := match_ast(graph, type_cast, "as_operator").get("__0__"))
    ):
        return build_binary_operation_node(args, "as", c_ids[0], typed_assignment)

    return build_expression_statement_node(
        args,
        c_ids=(_id for _id in c_ids if graph.nodes[_id]["label_type"] not in invalid_childs),
    )



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_group_d,
)
from blends.syntax.builders.file import (
    build_file_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    ignored_labels = {
        "\n",
        "\r\n",
    }
    graph = args.ast_graph
    c_ids = adj_ast(graph, args.n_id)

    filtered_ids = []
    for _id in c_ids:
        if (label_type := graph.nodes[_id]["label_type"]) in ignored_labels:
            continue
        if label_type == "import_declaration":
            filtered_ids.extend(match_ast_group_d(args.ast_graph, _id, "import_spec", depth=2))
            continue
        filtered_ids.append(_id)

    return build_file_node(args, filtered_ids)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    pred_ast,
)
from blends.syntax.builders.pair import (
    build_pair_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    p_id = pred_ast(graph, args.n_id)[0]
    c_ids = [_id for _id in adj_ast(graph, p_id) if _id != args.n_id]

    key_id = c_ids[0]
    value_id = c_ids[-1]

    return build_pair_node(args, key_id, value_id)



from blends.models import (
    NId,
)
from blends.stack.attributes import (
    STACK_GRAPH_IS_REFERENCE,
)
from blends.stack.edges import (
    Edge,
    add_edge,
)
from blends.stack.node_helpers import (
    pop_scoped_symbol_node_attributes,
    root_node_attributes,
)
from blends.stack.policies.registry import (
    get_export_policy,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)

_PATTERN_NODE_TYPES: frozenset[str] = frozenset(
    {"pattern_list", "tuple_pattern", "list_pattern", "list_splat_pattern"}
)


def extract_pattern_identifier_symbols(
    args: SyntaxGraphArgs, *, pattern_id: NId
) -> list[tuple[NId, str]]:
    label_type = args.ast_graph.nodes.get(pattern_id, {}).get("label_type")
    if label_type == "identifier":
        symbol = node_to_str(args.ast_graph, pattern_id)
        return [(pattern_id, symbol)] if symbol else []
    if label_type in _PATTERN_NODE_TYPES:
        return [
            result
            for child_id in args.ast_graph.successors(pattern_id)
            for result in extract_pattern_identifier_symbols(args, pattern_id=child_id)
        ]
    return []


def add_pattern_binding_nodes(args: SyntaxGraphArgs, *, var_id: NId) -> None:
    scope_stack = args.metadata.setdefault("scope_stack", [])
    parent_scope = scope_stack[-1] if scope_stack else None
    for identifier_nid, symbol in extract_pattern_identifier_symbols(args, pattern_id=var_id):
        synthetic_nid = get_next_synthetic_node_id(args)
        precedence = get_next_binding_precedence(args)
        args.syntax_graph.add_node(
            synthetic_nid,
            label_type="SyntheticPatternBinding",
            **pop_scoped_symbol_node_attributes(symbol=symbol, precedence=precedence),
        )
        if parent_scope:
            add_edge(
                args.syntax_graph,
                Edge(source=parent_scope, sink=synthetic_nid, precedence=precedence),
            )
        if args.syntax_graph.nodes.get(identifier_nid):
            args.syntax_graph.update_node(identifier_nid, {STACK_GRAPH_IS_REFERENCE: False})


def bound_identifier_symbol(args: SyntaxGraphArgs, *, var_id: NId) -> str | None:
    var_node = args.ast_graph.nodes.get(var_id, {})
    if var_node.get("label_type") != "identifier":
        return None
    symbol = node_to_str(args.ast_graph, var_id)
    return symbol if symbol else None


def get_next_binding_precedence(args: SyntaxGraphArgs) -> int:
    scope_stack = args.metadata.get("scope_stack", [])
    if not scope_stack:
        return 0
    return 1


def is_definition_exported(args: SyntaxGraphArgs) -> bool:
    policy = get_export_policy(args.language)
    scope_stack = args.metadata.get("scope_stack", [])
    parent_scope = scope_stack[-1] if scope_stack else None
    return policy.is_definition_exported(args.syntax_graph, args.n_id, parent_scope)


def get_next_synthetic_node_id(args: SyntaxGraphArgs) -> NId:
    current_id = args.metadata.get("next_synthetic_id")
    if current_id is None:
        msg = "next_synthetic_id not initialized in metadata"
        raise ValueError(msg)
    args.metadata["next_synthetic_id"] = current_id + 1
    return NId(current_id)


def get_or_create_root_nid(args: SyntaxGraphArgs) -> NId:
    existing = args.metadata.get("root_nid")
    if existing is not None:
        return existing
    root_nid = get_next_synthetic_node_id(args)
    args.syntax_graph.add_node(root_nid, **root_node_attributes())
    args.metadata["root_nid"] = root_nid
    return root_nid



from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_spread_element_node(args: SyntaxGraphArgs, identifier_id: NId) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        value_id=identifier_id,
        label_type="SpreadElement",
    )

    args.syntax_graph.add_edge(
        args.n_id,
        args.generic(args.fork_n_id(identifier_id)),
        label_ast="AST",
    )

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    childs = match_ast(graph, args.n_id, "del", "global")
    expr_id = childs.get("__0__")
    if not expr_id:
        expr_id = adj_ast(graph, args.n_id)[0]
    return args.generic(args.fork_n_id(expr_id))



from blends.symbolic_evaluation.dispatcher import (
    generic,
)
from blends.symbolic_evaluation.models import (
    SymbolicEvalArgs,
    SymbolicEvaluation,
)


def evaluate(args: SymbolicEvalArgs) -> SymbolicEvaluation:
    n_attr = args.graph.nodes[args.n_id]
    key_danger = generic(args.fork_n_id(n_attr["key_id"])).danger
    val_danger = generic(args.fork_n_id(n_attr["value_id"])).danger

    args.evaluation[args.n_id] = key_danger or val_danger

    if args.method_evaluators and (method_evaluator := args.method_evaluators.get("pair")):
        args.evaluation[args.n_id] = method_evaluator(args).danger

    return SymbolicEvaluation(args.evaluation[args.n_id], args.triggers)



from blends.syntax.models import (
    Dispatcher,
)
from blends.syntax.readers.kotlin import (
    annotation as kotlin_annotation,
)
from blends.syntax.readers.kotlin import (
    argument_list as kotlin_argument_list,
)
from blends.syntax.readers.kotlin import (
    assignment_expression as kotlin_assignment_expression,
)
from blends.syntax.readers.kotlin import (
    binary_expression as kotlin_binary_expression,
)
from blends.syntax.readers.kotlin import (
    catch_clause as kotlin_catch_clause,
)
from blends.syntax.readers.kotlin import (
    class_body as kotlin_class_body,
)
from blends.syntax.readers.kotlin import (
    class_declaration as kotlin_class_declaration,
)
from blends.syntax.readers.kotlin import (
    comment as kotlin_comment,
)
from blends.syntax.readers.kotlin import (
    companion_object as kotlin_companion_object,
)
from blends.syntax.readers.kotlin import (
    expression_statement as kotlin_expression_statement,
)
from blends.syntax.readers.kotlin import (
    finally_clause as kotlin_finally_clause,
)
from blends.syntax.readers.kotlin import (
    for_statement as kotlin_for_statement,
)
from blends.syntax.readers.kotlin import (
    identifier as kotlin_identifier,
)
from blends.syntax.readers.kotlin import (
    if_statement as kotlin_if_statement,
)
from blends.syntax.readers.kotlin import (
    import_declaration as kotlin_import_declaration,
)
from blends.syntax.readers.kotlin import (
    index_expression as kotlin_index_expression,
)
from blends.syntax.readers.kotlin import (
    infix_expression as kotlin_infix_expression,
)
from blends.syntax.readers.kotlin import (
    jump_statement as kotlin_jump_statement,
)
from blends.syntax.readers.kotlin import (
    member_access_expression as kotlin_member_access_expression,
)
from blends.syntax.readers.kotlin import (
    method_declaration as kotlin_method_declaration,
)
from blends.syntax.readers.kotlin import (
    method_invocation as kotlin_method_invocation,
)
from blends.syntax.readers.kotlin import (
    modifiers as kotlin_modifiers,
)
from blends.syntax.readers.kotlin import (
    number_literal as kotlin_number_literal,
)
from blends.syntax.readers.kotlin import (
    object as kotlin_object,
)
from blends.syntax.readers.kotlin import (
    parameter as kotlin_parameter,
)
from blends.syntax.readers.kotlin import (
    parameter_list as kotlin_parameter_list,
)
from blends.syntax.readers.kotlin import (
    parenthesized_expression as kotlin_parenthesized_expression,
)
from blends.syntax.readers.kotlin import (
    reserved_word as kotlin_reserved_word,
)
from blends.syntax.readers.kotlin import (
    source_file as kotlin_source_file,
)
from blends.syntax.readers.kotlin import (
    statement_block as kotlin_declaration_block,
)
from blends.syntax.readers.kotlin import (
    statements as kotlin_statements,
)
from blends.syntax.readers.kotlin import (
    string_content as kotlin_string_content,
)
from blends.syntax.readers.kotlin import (
    string_literal as kotlin_string_literal,
)
from blends.syntax.readers.kotlin import (
    try_statement as kotlin_try_statement,
)
from blends.syntax.readers.kotlin import (
    unary_expression as kotlin_unary_expression,
)
from blends.syntax.readers.kotlin import (
    variable_declaration as kotlin_variable_declaration,
)
from blends.syntax.readers.kotlin import (
    when_entry as kotlin_when_entry,
)
from blends.syntax.readers.kotlin import (
    when_expression as kotlin_when_expression,
)
from blends.syntax.readers.kotlin import (
    while_statement as kotlin_while_statement,
)

KOTLIN_DISPATCHERS: Dispatcher = {
    "annotation": kotlin_annotation.reader,
    "anonymous_initializer": kotlin_method_declaration.reader,
    "assignment": kotlin_assignment_expression.reader,
    "binary_expression": kotlin_binary_expression.reader,
    "block": kotlin_declaration_block.reader,
    "call_expression": kotlin_method_invocation.reader,
    "catch_block": kotlin_catch_clause.reader,
    "class_body": kotlin_class_body.reader,
    "class_declaration": kotlin_class_declaration.reader,
    "companion_object": kotlin_companion_object.reader,
    "expression": kotlin_expression_statement.reader,
    "function_body": kotlin_declaration_block.reader,
    "function_value_parameters": kotlin_parameter_list.reader,
    "finally_block": kotlin_finally_clause.reader,
    "for_statement": kotlin_for_statement.reader,
    "function_declaration": kotlin_method_declaration.reader,
    "identifier": kotlin_identifier.reader,
    "in_expression": kotlin_infix_expression.reader,
    "index_expression": kotlin_index_expression.reader,
    "infix_expression": kotlin_infix_expression.reader,
    "if_expression": kotlin_if_statement.reader,
    "import": kotlin_import_declaration.reader,
    "jump_expression": kotlin_jump_statement.reader,
    "lambda_literal": kotlin_declaration_block.reader,
    "block_comment": kotlin_comment.reader,
    "line_comment": kotlin_comment.reader,
    "modifiers": kotlin_modifiers.reader,
    "multiline_comment": kotlin_comment.reader,
    "navigation_expression": kotlin_member_access_expression.reader,
    "null": kotlin_reserved_word.reader,
    "number_literal": kotlin_number_literal.reader,
    "object_declaration": kotlin_class_declaration.reader,
    "object_literal": kotlin_object.reader,
    "package_header": kotlin_import_declaration.reader,
    "parameter": kotlin_parameter.reader,
    "parenthesized_expression": kotlin_parenthesized_expression.reader,
    "property_declaration": kotlin_variable_declaration.reader,
    "range_expression": kotlin_assignment_expression.reader,
    "return_expression": kotlin_jump_statement.reader,
    "source_file": kotlin_source_file.reader,
    "simple_identifier": kotlin_identifier.reader,
    "super_expression": kotlin_reserved_word.reader,
    "statement": kotlin_statements.reader,
    "string_literal": kotlin_string_literal.reader,
    "multiline_string_literal": kotlin_string_literal.reader,
    "string_content": kotlin_string_content.reader,
    "this_expression": kotlin_reserved_word.reader,
    "throw_expression": kotlin_jump_statement.reader,
    "try_expression": kotlin_try_statement.reader,
    "type_identifier": kotlin_identifier.reader,
    "unary_expression": kotlin_unary_expression.reader,
    "value_argument": kotlin_parameter.reader,
    "value_arguments": kotlin_argument_list.reader,
    "when_entry": kotlin_when_entry.reader,
    "when_expression": kotlin_when_expression.reader,
    "while_statement": kotlin_while_statement.reader,
}



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
)
from blends.syntax.builders.argument import (
    build_argument_node,
)
from blends.syntax.builders.named_argument import (
    build_named_argument_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.syntax.readers.constants import (
    C_SHARP_EXPRESSION,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    match = match_ast(graph, args.n_id, ":")
    if (
        match
        and (var_id := graph.nodes[args.n_id].get("label_field_name"))
        and (val_id := match.get("__1__"))
    ):
        arg_name = node_to_str(graph, var_id)
        return build_named_argument_node(args, arg_name, val_id)

    valid_childs = [
        _id
        for _id in adj_ast(graph, args.n_id)
        if graph.nodes[_id]["label_type"] in C_SHARP_EXPRESSION.union({"declaration_expression"})
    ]

    return build_argument_node(args, iter(valid_childs))



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_d,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    var_id = match_ast_d(graph, args.n_id, "val_definition")
    if not var_id:
        var_id = args.n_id

    var_type = None
    if var_type_id := graph.nodes[var_id].get("label_field_type"):
        var_type = graph.nodes[var_type_id].get("label_text")

    var_name_id = graph.nodes[var_id]["label_field_pattern"]
    var_name = node_to_str(graph, var_name_id)

    value_id = graph.nodes[var_id]["label_field_value"]

    return build_variable_declaration_node(args, var_name, var_type, value_id)



import re

from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_group_d,
)
from blends.syntax.builders.string_literal import (
    build_string_literal_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    literal_text = n_attrs.get("label_text") or node_to_str(graph, args.n_id)

    ints = match_ast_group_d(graph, args.n_id, "interpolation")
    int_subs = [adj_ast(graph, _id)[1] for _id in ints]

    vars_content = list(adj_ast(graph, args.n_id))
    for index, node_id in enumerate(vars_content):
        node = graph.nodes[node_id]
        if (
            node.get("label_type") != "string_content"
            or node.get("label_text") != "$"
            or (index + 1) >= len(vars_content)
        ):
            continue
        if (
            (next_node := vars_content[index + 1])
            and (label_text := graph.nodes[next_node].get("label_text"))
            and re.match(r"^[a-zA-Z_]", label_text)
        ):
            int_subs.append(next_node)

    if len(int_subs) > 0:
        return build_string_literal_node(args, literal_text, iter(int_subs))

    return build_string_literal_node(args, literal_text)



from blends.models import (
    NId,
)
from blends.syntax.builders.method_declaration import (
    DirectChildren,
    build_method_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    method = args.ast_graph.nodes[args.n_id]
    name_id = method["label_field_name"]
    name = node_to_str(args.ast_graph, name_id)
    block_id = method["label_field_body"]

    direct: DirectChildren = {
        "block_id": block_id,
        "parameters_id": method["label_field_parameters"],
    }

    return build_method_declaration_node(args, name, direct, {})



from blends.models import (
    NId,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    attrs = args.ast_graph.nodes[args.n_id]
    var_id = attrs["label_field_name"]
    val_id = attrs.get("label_field_type")
    return build_variable_declaration_node(args, var_id, None, val_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_group_d,
)
from blends.syntax.builders.argument_list import (
    build_argument_list_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    arg_ids = []

    for _id in adj_ast(graph, args.n_id):
        if graph.nodes[_id]["label_type"] == "value_arguments":
            arg_ids.extend(match_ast_group_d(graph, _id, "value_argument"))
        else:
            arg_ids.append(_id)

    return build_argument_list_node(args, iter(arg_ids))



from collections.abc import (
    Iterator,
)

from blends.symbolic_evaluation.context.search.model import (
    SearchArgs,
    SearchResult,
)


def search(args: SearchArgs) -> Iterator[SearchResult]:
    variable_id = args.graph.nodes[args.n_id]["variable_id"]
    if args.symbol == args.graph.nodes[variable_id].get("symbol"):
        yield False, args.n_id



from collections.abc import (
    Iterator,
)

from blends.symbolic_evaluation.context.search.model import (
    SearchArgs,
    SearchResult,
)


def search(args: SearchArgs) -> Iterator[SearchResult]:
    if not args.def_only:
        c_id = args.graph.nodes[args.n_id]["condition_id"]
        c_type = args.graph.nodes[c_id]["label_type"]
        if (c_type == "SymbolLookup" and args.symbol == args.graph.nodes[c_id]["symbol"]) or (
            c_type == "MethodInvocation"
            and (m_expr := args.graph.nodes[c_id].get("expression"))
            and args.symbol == str(m_expr).rsplit(".", 1)[0]
        ):
            yield False, c_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.if_statement import (
    build_if_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]

    if first_selector := match_ast_d(graph, args.n_id, "selector"):
        condition_id = first_selector
    else:
        child_ids = [
            c_id
            for c_id in adj_ast(graph, args.n_id)
            if graph.nodes[c_id].get("label_type")
            not in {
                "if",
                "(",
                ")",
                "else",
            }
        ]
        condition_id = child_ids[0]
    true_id = n_attrs["label_field_consequence"]
    false_id = n_attrs.get("label_field_alternative")
    return build_if_node(args, condition_id, true_id, false_id)



from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class ScopeStackNode:
    scope_index: int
    tail: ScopeStackNode | None


@dataclass(frozen=True, slots=True)
class SymbolStackNode:
    symbol_id: int
    scopes: ScopeStackNode | None
    tail: SymbolStackNode | None


@dataclass(frozen=True, slots=True)
class StackState:
    symbol_stack: SymbolStackNode | None
    scope_stack: ScopeStackNode | None



import re
from collections import (
    ChainMap,
)

from blends.ctx import ctx
from blends.models import (
    GraphDB,
    Language,
    NId,
    SyntaxGraph,
)
from blends.query import (
    adj_ast,
    match_ast_group_d,
)
from blends.symbolic_evaluation.context.search import (
    definition_search,
)
from blends.symbolic_evaluation.dispatcher import (
    generic,
)
from blends.symbolic_evaluation.models import (
    SymbolicEvalArgs,
)
from blends.symbolic_evaluation.utils import (
    get_backward_paths,
    get_current_class,
)


def is_symbol_literal(graph: SyntaxGraph, n_id: NId) -> bool:
    searched_symbol = graph.nodes[n_id].get("symbol")

    if not searched_symbol:
        return False

    var_values = set()
    for path in get_backward_paths(graph, n_id):
        if not (
            (var_def_id := definition_search(graph, path, searched_symbol))
            and (val_id := graph.nodes[var_def_id].get("value_id"))
        ):
            return False

        if graph.nodes[val_id]["label_type"] == "Literal":
            var_values.add(val_id)
        elif (
            graph.nodes[val_id]["label_type"] == "MethodInvocation"
            and (al_id := graph.nodes[val_id].get("arguments_id"))
            and (args_ids := adj_ast(graph, al_id))
            and is_symbol_literal(graph, args_ids[0])
        ):
            var_values.add(args_ids[0])
        else:
            return False

    return len(var_values) == 1


def method_returns_literals(graph: SyntaxGraph, method_id: NId) -> bool:
    return_nodes = match_ast_group_d(graph, method_id, "Return", -1)
    if all(
        (val_id := graph.nodes[_id].get("value_id"))
        and graph.nodes[val_id]["label_type"] == "Literal"
        for _id in return_nodes
    ):
        return True

    if (  # noqa: SIM103
        len(return_nodes) == 1
        and (val_id := graph.nodes[return_nodes[0]].get("value_id"))
        and is_symbol_literal(graph, val_id)
    ):
        return True

    return False


def evaluate_file(
    args: SymbolicEvalArgs,
    obj_construct: str,
    file: str,
    struct: dict,  # type: ignore [type-arg]
) -> tuple[bool, bool]:
    danger = False
    n_attrs = args.graph.nodes[args.n_id]
    if args.graph_db and obj_construct in struct:
        graph = shard.syntax_graph if (shard := args.graph_db.get_path_shard(file)) else None
        if not graph:
            return (danger, True)
        class_data = struct[obj_construct]["data"]
        method_id = class_data[n_attrs["expression"]]["node"]
        if return_nodes := match_ast_group_d(graph, method_id, "Return", -1):
            danger = any(
                generic(args.fork_n_id(node, path, graph)).danger
                for node in return_nodes
                for path in get_backward_paths(graph, node)
            )

        if method_returns_literals(graph, method_id):
            return (danger, False)
    return (danger, True)


def evaluate_method_extern_file(
    args: SymbolicEvalArgs,
    lang: Language,
    obj_instance: dict,  # type: ignore [type-arg]
    graph_db: GraphDB,
) -> tuple[bool, bool]:
    context_lang = next(
        (ctx_lang for ctx_lang in graph_db.context if ctx_lang.value == lang.value), None
    )
    if not context_lang:
        return (False, True)
    lang_context = graph_db.context[context_lang]
    obj_source = obj_instance["source"]

    working_dir = f"{ctx.working_dir}/"
    if obj_source.startswith(working_dir):
        obj_source = re.sub(working_dir, "", obj_source)

    if obj_instance["source_type"] == "package" and (
        source_context := lang_context.get(obj_source)
    ):
        for file, struct in source_context.items():
            return evaluate_file(
                args,
                obj_instance["object"],
                file,
                struct,
            )
    else:
        file_structures = dict(ChainMap(*lang_context.values()))
        if struct := file_structures.get(obj_source):
            return evaluate_file(
                args,
                obj_instance["object"],
                obj_source,
                struct,
            )
    return (False, True)


def evaluate_method_invocation(
    args: SymbolicEvalArgs,
    lang: Language,
) -> tuple[bool, bool]:
    graph = args.graph
    n_attrs = graph.nodes[args.n_id]
    metadata_node = graph.nodes[NId(0)]
    current_class = graph.nodes[get_current_class(graph, args.n_id)]["name"]
    if not args.graph_db:
        return (False, True)

    if not (obj_id := n_attrs.get("object_id")):
        if (
            (expr_method := n_attrs.get("expression"))
            and (class_methods := metadata_node["structure"].get(current_class, {}).get("data"))
            and (invocation_id := class_methods.get(expr_method, {}).get("node"))
        ):
            danger = any(
                generic(args.fork_n_id(node, path, graph)).danger
                for node in match_ast_group_d(graph, invocation_id, "Return", -1)
                for path in get_backward_paths(graph, node)
            )
            return (danger, not method_returns_literals(graph, invocation_id))
        return (False, True)

    if (
        (class_name := graph.nodes[obj_id].get("name"))
        and (
            class_methods := metadata_node["structure"]
            .get(current_class, {})
            .get("data", {})
            .get(class_name, {})
            .get("data")
        )
        and (expr_method := n_attrs.get("expression"))
        and (invocation_id := class_methods.get(expr_method, {}).get("node"))
    ):
        danger = any(
            generic(args.fork_n_id(node, path, graph)).danger
            for node in match_ast_group_d(graph, invocation_id, "Return", -1)
            for path in get_backward_paths(graph, node)
        )
        return (danger, not method_returns_literals(graph, invocation_id))

    if (
        (object_method := graph.nodes[obj_id].get("symbol"))
        and (class_instances := metadata_node["instances"].get(current_class))
        and (obj_instance := class_instances.get(object_method))
    ):
        return evaluate_method_extern_file(args, lang, obj_instance, args.graph_db)

    return (False, True)



from blends.models import (
    NId,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    value_id = n_attrs.get("label_field_value")
    name_id = n_attrs.get("label_field_name")
    var_name = node_to_str(graph, name_id) if name_id else node_to_str(graph, args.n_id)
    type_id = n_attrs.get("label_field_type")
    var_type = node_to_str(graph, type_id) if type_id else None

    return build_variable_declaration_node(args, var_name, var_type, value_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.return_statement import (
    build_return_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    ignored_labels = {"return", ";"}
    val_id: NId | None = None
    c_ids = [
        _id
        for _id in adj_ast(graph, args.n_id)
        if graph.nodes[_id].get("label_type") not in ignored_labels
    ]
    if c_ids:
        val_id = c_ids[0]

    return build_return_node(args, val_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
    match_ast_d,
)
from blends.syntax.builders.method_invocation import (
    DirectChildren,
    build_method_invocation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    nodes = graph.nodes

    if nodes[args.n_id]["label_type"] == "print_intrinsic":
        childs = adj_ast(graph, args.n_id)
        if (
            len(childs) > 1
            and (child_id := childs[1])
            and nodes[child_id]["label_type"] == "parenthesized_expression"
            and (c_ids := match_ast(graph, child_id))
        ):
            expr_id = c_ids["__1__"]
        else:
            expr_id = None
        children: DirectChildren = {}
        if expr_id:
            children["expression_id"] = expr_id
        return build_method_invocation_node(args, "print", children)

    if nodes[args.n_id]["label_type"] == "exit_statement":
        args_id = match_ast(graph, args.n_id).get("__2__", None)
        children = {}
        if args_id:
            children["arguments_id"] = args_id
        return build_method_invocation_node(args, "exit", children)

    args_id = match_ast_d(args.ast_graph, args.n_id, "echo")
    children = {}
    if args_id:
        children["arguments_id"] = args_id
    return build_method_invocation_node(args, "echo", children)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    pred_ast,
)
from blends.syntax.models import (
    SyntaxCfgArgs,
)


def build(args: SyntaxCfgArgs) -> NId:
    childs = adj_ast(args.graph, args.n_id)
    if len(childs) == 0:
        parent_id = pred_ast(args.graph, args.n_id)[0]
        siblings = adj_ast(args.graph, parent_id)
        next_sibling = next(
            (_id for _id in siblings if len(adj_ast(args.graph, _id)) > 0 and _id > args.n_id),
            siblings[-1],
        )
        childs = adj_ast(args.graph, next_sibling)
    for c_id in childs:
        args.graph.add_edge(
            args.n_id,
            args.generic(args.fork(c_id, args.nxt_id)),
            label_cfg="CFG",
        )
    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
)
from blends.syntax.builders.assignment import (
    build_assignment_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attr = graph.nodes[args.n_id]
    var_id = n_attr.get("label_field_left", "")
    val_id = n_attr.get("label_field_right", "")

    operator = None
    if op_id := n_attr.get("label_field_operator"):
        operator = graph.nodes[op_id]["label_text"]

    if (
        graph.nodes[var_id]["label_type"] == "assignable_expression"
        and len(adj_ast(graph, var_id)) == 1
        and (var_n_id := match_ast(graph, var_id).get("__0__"))
    ):
        var_id = var_n_id

    return build_assignment_node(args, var_id, val_id, operator)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
    match_ast_d,
)
from blends.syntax.builders.assignment import (
    build_assignment_node,
)
from blends.syntax.builders.expression_statement import (
    build_expression_statement_node,
)
from blends.syntax.builders.method_invocation import (
    DirectChildren,
    build_method_invocation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def _try_build_assignment(
    args: SyntaxGraphArgs,
    filtered_ids: list[NId],
) -> NId | None:
    graph = args.ast_graph
    if not (
        len(filtered_ids) == 1
        and graph.nodes[filtered_ids[0]]["label_type"] == "assignment_expression"
    ):
        return None
    n_attr: dict[str, str] = graph.nodes[filtered_ids[0]]
    var_id = NId(n_attr["label_field_left"])
    op_id = n_attr.get("label_field_operator")
    if (
        graph.nodes[var_id]["label_type"] == "assignable_expression"
        and len(adj_ast(graph, var_id)) == 1
        and (var_n_id := match_ast(graph, var_id).get("__0__"))
    ):
        var_id = var_n_id
    if op_id:
        return build_assignment_node(
            args,
            NId(var_id),
            NId(n_attr["label_field_right"]),
            graph.nodes[op_id]["label_text"],
        )
    return None


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    c_ids = adj_ast(args.ast_graph, args.n_id)
    ignored_types = {";", "(", ")", "super", "comment"}
    filtered_ids = [
        _id for _id in c_ids if args.ast_graph.nodes[_id]["label_type"] not in ignored_types
    ]
    expected_filtered_children_for_method_invocation = 2
    if (
        len(filtered_ids) == expected_filtered_children_for_method_invocation
        and args.ast_graph.nodes[filtered_ids[0]]["label_type"] == "identifier"
        and args.ast_graph.nodes[filtered_ids[1]]["label_type"] == "selector"
        and (select_1 := match_ast_d(graph, filtered_ids[1], "argument_part"))
    ):
        expr = args.ast_graph.nodes[filtered_ids[0]]["label_text"]
        args_id = match_ast_d(graph, select_1, "arguments")
        children: DirectChildren = {}
        if args_id:
            children["arguments_id"] = args_id
        return build_method_invocation_node(args, expr, children)

    expected_selectors_for_chained_method_invocation = 2
    if (
        (sel := adj_ast(graph, args.n_id, label_type="selector"))
        and len(sel) == expected_selectors_for_chained_method_invocation
        and (select_1 := match_ast_d(graph, sel[0], "unconditional_assignable_selector"))
        and (select_2 := match_ast_d(graph, sel[1], "argument_part"))
    ):
        child_s1 = adj_ast(graph, select_1)
        expr_id = child_s1[1] if len(child_s1) > 1 else child_s1[0]
        expr = node_to_str(graph, filtered_ids[0]) + "." + node_to_str(graph, expr_id)
        args_id = match_ast_d(graph, select_2, "arguments")
        children = {"expression_id": expr_id, "object_id": filtered_ids[0]}
        if args_id:
            children["arguments_id"] = args_id
        return build_method_invocation_node(args, expr, children)

    if assignment := _try_build_assignment(args, filtered_ids):
        return assignment

    return build_expression_statement_node(args, iter(filtered_ids))



def split_on_last_dot(string: str) -> tuple[str, str]:
    portions = string.rsplit(".", maxsplit=1)
    expected_portions_after_split = 2
    if len(portions) == expected_portions_after_split:
        return portions[0], portions[1]
    return portions[0], ""



from blends.ctx import ctx
from blends.models import (
    NId,
)
from blends.stack.edges import (
    Edge,
    add_edge,
)
from blends.stack.node_helpers import (
    push_symbol_node_attributes,
)
from blends.syntax.builders.utils import (
    get_next_synthetic_node_id,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_member_access_node(
    args: SyntaxGraphArgs,
    member: str,
    expression: str,
    expression_id: NId,
) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        member=member,
        expression=expression,
        expression_id=expression_id,
        label_type="MemberAccess",
    )

    child_id = args.generic(args.fork_n_id(expression_id))
    args.syntax_graph.add_edge(
        args.n_id,
        child_id,
        label_ast="AST",
    )

    if ctx.has_feature_flag("StackGraph"):
        args.syntax_graph.update_node(
            args.n_id,
            push_symbol_node_attributes(symbol=member, is_reference=True),
        )
        dot_id = get_next_synthetic_node_id(args)
        args.syntax_graph.add_node(
            dot_id,
            label_type="SyntheticMemberDot",
            **push_symbol_node_attributes(symbol=".", is_reference=False),
        )
        add_edge(args.syntax_graph, Edge(source=args.n_id, sink=dot_id))
        add_edge(args.syntax_graph, Edge(source=dot_id, sink=child_id))

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.binary_operation import (
    build_binary_operation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    nodes = graph.nodes
    operator_id = nodes[args.n_id]["label_field_operator"]
    operator = nodes[operator_id]["label_text"]
    left_id = nodes[args.n_id]["label_field_left"]
    right_id = nodes[args.n_id]["label_field_right"]

    if nodes[left_id].get("label_type") == "parenthesized_expression":
        c_ids = match_ast(graph, left_id)
        left_id = c_ids.get("__1__")
    if nodes[right_id].get("label_type") == "parenthesized_expression":
        c_ids = match_ast(graph, right_id)
        right_id = c_ids.get("__1__")

    return build_binary_operation_node(args, operator, left_id, right_id)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.pair import (
    build_pair_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    key_id = graph.nodes[args.n_id]["label_field_key"]
    value_id = graph.nodes[args.n_id]["label_field_value"]
    if graph.nodes[key_id]["label_type"] == "computed_property_name" and (
        c_id := match_ast(graph, key_id, "[", "]").get("__0__")
    ):
        key_id = c_id

    return build_pair_node(args, key_id, value_id)



import pytest

from test.stack.test_helpers import (
    count_defs_refs,
    find_symbol_nodes,
    resolve_symbol,
    setup_stack_resolution,
)


@pytest.mark.blends_test_group("stack_unittesting")
def test_tuple_unpacking() -> None:
    view, record = setup_stack_resolution("case_tuple_unpacking.py")
    for symbol in ["a", "b"]:
        defs, refs = count_defs_refs(view, symbol)
        assert defs == 1, f"Expected 1 def of '{symbol}', got {defs}"
        assert refs == 1, f"Expected 1 ref to '{symbol}', got {refs}"
        result = resolve_symbol(view, record, symbol)
        assert len(result) == 1, f"Expected '{symbol}' to resolve to 1 def, got {len(result)}"
        symbol_id = view.symbols.symbol_to_id.get(symbol)
        assert symbol_id is not None
        def_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=True)
        assert result == def_nodes, f"'{symbol}' should resolve to its tuple-unpacking definition"


@pytest.mark.blends_test_group("stack_unittesting")
def test_for_tuple_unpack() -> None:
    view, record = setup_stack_resolution("case_for_tuple_unpack.py")
    for symbol in ["a", "b"]:
        defs, refs = count_defs_refs(view, symbol)
        assert defs == 1, f"Expected 1 def of '{symbol}', got {defs}"
        assert refs == 1, f"Expected 1 ref to '{symbol}', got {refs}"
        result = resolve_symbol(view, record, symbol)
        assert len(result) == 1, f"Expected '{symbol}' to resolve to 1 def, got {len(result)}"
        symbol_id = view.symbols.symbol_to_id.get(symbol)
        assert symbol_id is not None
        def_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=True)
        assert result == def_nodes, f"'{symbol}' should resolve to its for-loop pattern definition"


@pytest.mark.blends_test_group("stack_unittesting")
def test_nested_unpacking() -> None:
    view, record = setup_stack_resolution("case_nested_unpacking.py")
    for symbol in ["a", "b", "c"]:
        defs, refs = count_defs_refs(view, symbol)
        assert defs == 1, f"Expected 1 def of '{symbol}', got {defs}"
        assert refs == 1, f"Expected 1 ref to '{symbol}', got {refs}"
        result = resolve_symbol(view, record, symbol)
        assert len(result) == 1, f"Expected '{symbol}' to resolve to 1 def, got {len(result)}"
        symbol_id = view.symbols.symbol_to_id.get(symbol)
        assert symbol_id is not None
        def_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=True)
        assert result == def_nodes, f"'{symbol}' should resolve to its nested-unpacking definition"


@pytest.mark.blends_test_group("stack_unittesting")
def test_star_unpacking() -> None:
    view, record = setup_stack_resolution("case_star_unpacking.py")
    for symbol in ["first", "rest", "last"]:
        defs, refs = count_defs_refs(view, symbol)
        assert defs == 1, f"Expected 1 def of '{symbol}', got {defs}"
        assert refs == 1, f"Expected 1 ref to '{symbol}', got {refs}"
        result = resolve_symbol(view, record, symbol)
        assert len(result) == 1, f"Expected '{symbol}' to resolve to 1 def, got {len(result)}"
        symbol_id = view.symbols.symbol_to_id.get(symbol)
        assert symbol_id is not None
        def_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=True)
        assert result == def_nodes, f"'{symbol}' should resolve to its star-unpacking definition"



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.argument import (
    build_argument_node,
)
from blends.syntax.builders.named_argument import (
    build_named_argument_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    childs = adj_ast(graph, args.n_id)
    n_attrs = graph.nodes[args.n_id]
    value_id = n_attrs.get("label_field_value")
    if not value_id:
        value_id = childs[-1]

    if identifier_id := n_attrs.get("label_field_name"):
        arg_name = node_to_str(graph, identifier_id)
        return build_named_argument_node(args, arg_name, value_id)

    return build_argument_node(args, iter(childs))



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
    match_ast_d,
    search_pred_until_type,
)
from blends.syntax.builders.method_declaration import (
    DirectChildren,
    build_method_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    body_parents = {
        "class_body",
        "extension_body",
        "lambda_expression",
        "program",
    }
    body_id = None

    name_id = graph.nodes[args.n_id]["label_field_name"]
    m_name = node_to_str(graph, name_id)
    parents = search_pred_until_type(graph, args.n_id, body_parents)

    direct: DirectChildren = {}
    if parents and (class_childs := list(adj_ast(graph, parents[0]))):
        pm_id = match_ast_d(graph, args.n_id, "formal_parameter_list")
        if pm_id and "__0__" in match_ast(args.ast_graph, pm_id, "(", ")"):
            direct["parameters_id"] = pm_id
        body_id = class_childs[class_childs.index(parents[1]) + 1]

    if body_id:
        direct["block_id"] = body_id

    return build_method_declaration_node(args, m_name, direct, {})



from blends.symbolic_evaluation.dispatcher import (
    generic,
)
from blends.symbolic_evaluation.models import (
    SymbolicEvalArgs,
    SymbolicEvaluation,
)


def evaluate(args: SymbolicEvalArgs) -> SymbolicEvaluation:
    danger_cond = False
    danger_true = False
    danger_false = False
    node = args.graph.nodes[args.n_id]
    if _id := node.get("condition_id"):
        danger_cond = generic(args.fork_n_id(_id)).danger
    if args.n_id not in args.path:
        if _id := node.get("true_id"):
            danger_true = generic(args.fork_n_id(_id)).danger
        if _id := node.get("false_id"):
            danger_false = generic(args.fork_n_id(_id)).danger

    args.evaluation[args.n_id] = danger_cond or danger_true or danger_false

    return SymbolicEvaluation(args.evaluation[args.n_id], args.triggers)



from __future__ import annotations

from typing import TYPE_CHECKING

from blends.stack.policies.base import ExportPolicy
from blends.stack.scope_types import ScopeType

if TYPE_CHECKING:
    from blends.models import NId, SyntaxGraph


class DefaultExportPolicy(ExportPolicy):
    def is_scope_exported(
        self,
        _graph: SyntaxGraph,
        _node_id: NId,
        scope_type: str | None,
    ) -> bool:
        return scope_type == ScopeType.FILE.value

    def is_definition_exported(
        self,
        graph: SyntaxGraph,
        _node_id: NId,
        parent_scope_id: NId | None,
    ) -> bool:
        if parent_scope_id is None:
            return True
        parent_type = graph.nodes[parent_scope_id].get("stack_graph_scope_type")
        return parent_type == ScopeType.FILE.value



import logging
from collections.abc import Iterator
from itertools import (
    count,
)
from pathlib import Path

from tree_sitter import Node

from blends.content.content import get_content_by_path
from blends.models import Content, Graph, Language, NId
from blends.tree import FIELDS_BY_LANGUAGE
from blends.tree.parse import ParsingError, get_tree

LOGGER = logging.getLogger(__name__)


def _hash_node(node: Node) -> int:
    return hash((node.end_point, node.start_point, node.type))


def _is_node_terminal(node: Node, language: Language) -> bool:
    return (
        language == Language.YAML
        and node.type
        in [
            "flow_mapping",
            "single_quote_scalar",
            "double_quote_scalar",
        ]
    ) or (language == Language.SWIFT and node.type == "bang")


def _has_content_node(node: Node, language: Language) -> bool:
    return (
        (
            language == Language.CSHARP
            and node.type
            in {
                "character_literal",
                "predefined_type",
                "string_literal",
                "verbatim_string_literal",
            }
        )
        or (
            language == Language.DART
            and node.type
            in {
                "list_literal",
                "set_or_map_literal",
                "string_literal",
            }
        )
        or (
            language == Language.GO
            and node.type
            in {
                "interpreted_string_literal",
            }
        )
        or (
            language
            in {
                Language.JAVASCRIPT,
                Language.TYPESCRIPT,
            }
            and node.type
            in {
                "template_string",
                "regex",
            }
        )
        or (
            language == Language.KOTLIN
            and node.type
            in {
                "string_literal",
            }
        )
        or (language == Language.SWIFT and node.type == "bang")
        or (
            language == Language.YAML
            and node.type
            in {
                "single_quote_scalar",
                "double_quote_scalar",
                "flow_mapping",
            }
        )
    )


def _build_ast_graph(
    content: Content,
    node: Node,
    counter: Iterator[str],
    graph: Graph,
    *,
    _edge_index: NId | None = None,
    _parent: NId | None = None,
    _parent_fields: dict[int, str] | None = None,
) -> Graph:
    if not isinstance(node, Node):
        raise NotImplementedError

    n_id = next(counter)
    raw_l, raw_c = node.start_point

    graph.add_node(n_id, label_l=str(raw_l + 1), label_c=str(raw_c + 1), label_type=node.type)

    if _parent is not None:
        graph.add_edge(_parent, n_id, label_ast="AST", label_index=str(_edge_index))

    if (field := (_parent_fields or {}).get(_hash_node(node))) and _parent is not None:
        graph.nodes[_parent][f"label_field_{field}"] = n_id

    if not node.children or _has_content_node(node, content.language):
        node_content = content.content_bytes[node.start_byte : node.end_byte]
        graph.nodes[n_id]["label_text"] = node_content.decode("latin-1")

    if node.children and not _is_node_terminal(node, content.language):
        for edge_index, child in enumerate(node.children):
            child_node: Node = child
            _build_ast_graph(
                content,
                child_node,
                counter,
                graph,
                _edge_index=NId(edge_index),
                _parent=n_id,
                _parent_fields={
                    _hash_node(child): fld
                    for fld in FIELDS_BY_LANGUAGE[content.language].get(node.type, ())
                    for child in [node.child_by_field_name(fld)]
                    if child
                },
            )

    return graph


def _get_ast_from_content(content: Content) -> Graph | None:
    try:
        tree = get_tree(content)
    except ParsingError:
        return None
    counter = map(str, count(1))
    try:
        return _build_ast_graph(content, tree.root_node, counter, Graph())
    except NotImplementedError:
        LOGGER.exception("Error building the ast_graph for %s", content.path)

    return None


def get_ast_graph_from_path(path: Path) -> Graph | None:
    if (content := get_content_by_path(path)) is None:
        LOGGER.error("Error extracting the content of %s", path)
        return None

    if content.language == Language.NOT_SUPPORTED:
        LOGGER.warning("Unsupported language in %s", path)
        return None

    return _get_ast_from_content(content)


def get_ast_graph_from_content(content: Content) -> Graph | None:
    if content.language == Language.NOT_SUPPORTED:
        LOGGER.warning("Unsupported language in %s", content.path)
        return None

    return _get_ast_from_content(content)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.file import (
    build_file_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    valid_types = {
        "class_declaration",
        "delegate_declaration",
        "enum_declaration",
        "extern_alias_directive",
        "identifier",
        "interface_declaration",
        "record_declaration",
        "record_struct_declaration",
        "struct_declaration",
        "using_directive",
    }
    childs_id = adj_ast(args.ast_graph, args.n_id)

    return build_file_node(
        args,
        c_ids=(c_id for c_id in childs_id if graph.nodes[c_id]["label_type"] in valid_types),
    )



from blends.ctx import ctx
from blends.models import (
    NId,
)
from blends.query import (
    pred_ast,
)
from blends.syntax.builders.import_module import (
    build_import_module_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    nodes = graph.nodes
    alias: str | None = None
    expression = ""

    import_clauses = [
        _id
        for _id in pred_ast(graph, args.n_id, -1)
        if nodes[_id].get("label_type") == "import_statement"
    ]

    module_path: str | None = None
    label_field_n_id = nodes[import_clauses[0]].get("label_field_source")
    if label_field_n_id:
        expression = node_to_str(graph, label_field_n_id)[1:-1]
        if ctx.has_feature_flag("StackGraph"):
            module_path = expression

    identifier_n_id = nodes[args.n_id].get("label_field_name")
    identifier = ""
    if identifier_n_id:
        identifier = nodes[identifier_n_id].get("label_text", "")
    expression += "." + identifier
    if alias_n_id := nodes[args.n_id].get("label_field_alias"):
        alias = nodes[alias_n_id].get("label_text")
    else:
        alias = identifier

    return build_import_module_node(args, expression, alias, module_path=module_path)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.declaration_block import (
    build_declaration_block_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    _, *c_ids, _ = adj_ast(args.ast_graph, args.n_id)  # do not consider { }

    return build_declaration_block_node(args, c_ids)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.method_invocation import (
    DirectChildren,
    build_method_invocation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    nodes = graph.nodes
    n_attrs = nodes[args.n_id]

    ignored_types = {"(", ",", ")"}

    expr_id = n_attrs.get("label_field_function") or n_attrs.get("label_field_object")
    expr = ""
    if expr_id is not None:
        if n_attrs.get("label_type") == "member_call_expression":
            raw_expr = node_to_str(graph, expr_id)
            method_name_n_id = n_attrs["label_field_name"]
            method_name = nodes[method_name_n_id].get("label_text") or node_to_str(
                graph,
                method_name_n_id,
            )
            exp_tokens = [_exp.split("(", maxsplit=1)[0] for _exp in raw_expr.split("->")]
            exp_tokens.append(method_name)
            expr = "->".join(exp_tokens)
        else:
            expr = node_to_str(graph, expr_id)

    if arguments_id := graph.nodes[args.n_id].get("label_field_arguments"):
        arguments = {
            _id
            for _id in adj_ast(graph, arguments_id)
            if graph.nodes[_id].get("label_type") not in ignored_types
        }
        if not arguments:
            arguments_id = None

    children: DirectChildren = {}
    if expr_id:
        children["expression_id"] = expr_id
    if arguments_id:
        children["arguments_id"] = arguments_id

    return build_method_invocation_node(args, expr, children)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_d,
    match_ast_group_d,
)
from blends.syntax.builders.parameter import (
    build_parameter_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    var_type = None
    var_name = None
    c_ids = match_ast_group_d(graph, args.n_id, "annotation")

    type_id = n_attrs.get("label_field_type") or match_ast_d(graph, args.n_id, "type_identifier")
    if type_id:
        var_type = node_to_str(graph, type_id)
    identifier_id = n_attrs.get("label_field_name") or match_ast_d(graph, args.n_id, "identifier")
    if identifier_id:
        var_name = node_to_str(graph, identifier_id)

    value_id = n_attrs.get("label_field_value")
    return build_parameter_node(
        args=args,
        variable=var_name,
        variable_type=var_type,
        value_id=value_id,
        c_ids=c_ids,
    )



from blends.models import (
    NId,
)
from blends.syntax.builders.method_declaration import (
    DirectChildren,
    build_method_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    as_attrs = args.ast_graph.nodes[args.n_id]
    method_name = (
        node_to_str(args.ast_graph, name_id)
        if (name_id := as_attrs.get("label_field_name"))
        else None
    )

    block_id = as_attrs["label_field_body"]
    class_id = as_attrs["label_field_class"]

    direct: DirectChildren = {"block_id": block_id, "modifiers_id": class_id}

    return build_method_declaration_node(args, method_name, direct, {})
